# coding: utf-8
# /*##########################################################################
# Copyright (C) 2016-2020 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################

"""Test compatibility with previously existing NexusPath classes."""


__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "10/02/2022"


import pytest

from nxtomo.paths.nxtomo import get_paths as new_get_paths
from nxtomo.paths.nxtomo import nx_tomo_path_latest


# classes which were previously defining path to save data as NXtomo from tomoscan.esrf.scan.nxtomoscan.py file
class _NEXUS_PATHS:
    """Register paths for NXtomo.
    The raw data are those of the initial version.
    If the value is None then the path did not exist originally.
    """

    PROJ_PATH = "instrument/detector/data"

    SCAN_META_PATH = "scan_meta/technique/scan"

    DET_META_PATH = "scan_meta/technique/detector"

    ROTATION_ANGLE_PATH = "sample/rotation_angle"

    SAMPLE_PATH = "sample"

    NAME_PATH = "sample/name"

    GRP_SIZE_ATTR = "group_size"

    SAMPLE_NAME_PATH = "sample/sample_name"

    X_TRANS_PATH = "sample/x_translation"

    Y_TRANS_PATH = "sample/y_translation"

    Z_TRANS_PATH = "sample/z_translation"

    IMG_KEY_PATH = "instrument/detector/image_key"

    IMG_KEY_CONTROL_PATH = "instrument/detector/image_key_control"

    X_PIXEL_SIZE_PATH = "instrument/detector/x_pixel_size"

    Y_PIXEL_SIZE_PATH = "instrument/detector/y_pixel_size"

    X_PIXEL_MAG_SIZE_PATH = "instrument/detector/x_magnified_pixel_size"

    Y_PIXEL_MAG_SIZE_PATH = "instrument/detector/y_magnified_pixel_size"

    DISTANCE_PATH = "instrument/detector/distance"

    FOV_PATH = "instrument/detector/field_of_view"

    EXPOSURE_TIME_PATH = "instrument/detector/count_time"

    TOMO_N_SCAN = "instrument/detector/tomo_n"

    ENERGY_PATH = "beam/incident_energy"

    START_TIME_PATH = "start_time"

    END_TIME_START = "end_time"  # typo - deprecated

    END_TIME_PATH = "end_time"

    INTENSITY_MONITOR_PATH = "diode/data"

    SOURCE_NAME = None

    SOURCE_TYPE = None

    SOURCE_PROBE = None

    INSTRUMENT_NAME = None


class _NEXUS_PATHS_V_1_0(_NEXUS_PATHS):
    pass


class _NEXUS_PATHS_V_1_1(_NEXUS_PATHS_V_1_0):
    ENERGY_PATH = "instrument/beam/incident_energy"

    SOURCE_NAME = "instrument/source/name"

    SOURCE_TYPE = "instrument/source/type"

    SOURCE_PROBE = "instrument/source/probe"

    INSTRUMENT_NAME = "instrument/name"

    NAME_PATH = "title"

    SAMPLE_NAME_PATH = "sample/name"


_class_to_compare_versions = {
    1.0: (_NEXUS_PATHS_V_1_0, new_get_paths(1.0)),
    1.1: (_NEXUS_PATHS_V_1_1, new_get_paths(1.1)),
}


@pytest.mark.parametrize("path_version", (1.0, 1.1))
def test_compare_result(path_version):
    """Ensure the new way of providing NeXus paths does not break the previous API or values."""
    old_class, new_class = _class_to_compare_versions[path_version]
    assert old_class.PROJ_PATH == new_class.PROJ_PATH
    assert old_class.SCAN_META_PATH == new_class.SCAN_META_PATH
    assert old_class.DET_META_PATH == new_class.DET_META_PATH
    assert old_class.ROTATION_ANGLE_PATH == new_class.ROTATION_ANGLE_PATH
    assert old_class.SAMPLE_PATH == new_class.SAMPLE_PATH
    assert old_class.NAME_PATH == new_class.NAME_PATH
    assert old_class.GRP_SIZE_ATTR == new_class.GRP_SIZE_ATTR
    assert old_class.SAMPLE_NAME_PATH == new_class.SAMPLE_NAME_PATH
    assert old_class.X_TRANS_PATH == new_class.X_TRANS_PATH
    assert old_class.Y_TRANS_PATH == new_class.Y_TRANS_PATH
    assert old_class.Z_TRANS_PATH == new_class.Z_TRANS_PATH
    assert old_class.IMG_KEY_PATH == new_class.IMG_KEY_PATH
    assert old_class.IMG_KEY_CONTROL_PATH == new_class.IMG_KEY_CONTROL_PATH
    assert old_class.X_PIXEL_SIZE_PATH == new_class.X_PIXEL_SIZE_PATH
    assert old_class.Y_PIXEL_SIZE_PATH == new_class.Y_PIXEL_SIZE_PATH
    assert old_class.DISTANCE_PATH == new_class.DISTANCE_PATH
    assert old_class.FOV_PATH == new_class.FOV_PATH
    assert old_class.EXPOSURE_TIME_PATH == new_class.EXPOSURE_TIME_PATH
    assert old_class.TOMO_N_SCAN == new_class.TOMO_N_SCAN
    assert old_class.ENERGY_PATH == new_class.ENERGY_PATH
    assert old_class.START_TIME_PATH == new_class.START_TIME_PATH
    assert old_class.END_TIME_PATH == new_class.END_TIME_PATH
    assert old_class.INTENSITY_MONITOR_PATH == new_class.INTENSITY_MONITOR_PATH
    assert old_class.SOURCE_NAME == new_class.SOURCE_NAME
    assert old_class.SOURCE_TYPE == new_class.SOURCE_TYPE
    assert old_class.INSTRUMENT_NAME == new_class.INSTRUMENT_NAME


def test_unknow_nexus_path_version():
    assert new_get_paths(None) == nx_tomo_path_latest
    assert new_get_paths(1.99) == nx_tomo_path_latest
    with pytest.raises(ValueError):
        assert new_get_paths(-1.0) is None
    with pytest.raises(ValueError):
        assert new_get_paths(999.0) is None



"""
Internal decorator utilities.
"""

from functools import wraps

import pint

# Initialize the unit registry
ureg = pint.UnitRegistry()


def check_dimensionality(expected_dimension: str, allow_none: bool = True):
    """
    Decorator to check the dimensionality of a pint.Quantity parameter.

    :param expected_dimension: expected dimensionality of the parameter, e.g. "[length]".
    """

    def check_parameter(parameter, parameter_name: str, allow_none: bool):
        # Check if the value is a pint.Quantity and has the expected dimensionality
        if isinstance(parameter, pint.Quantity):
            if parameter.dimensionality != ureg.get_dimensionality(expected_dimension):
                raise TypeError(
                    f"{parameter_name}: expected dimensionality {expected_dimension}, but got {parameter.dimensionality}."
                )
        elif allow_none:
            if parameter is not None:
                raise TypeError(
                    f"{parameter_name} must be a pint.Quantity or None. Got {type(parameter)}."
                )
        else:
            raise TypeError(
                f"{parameter_name} must be a pint.Quantity. Got {type(parameter)}."
            )

    def decorator(func):
        @wraps(func)
        def wrapper(self, value):
            check_parameter(
                parameter=value,
                parameter_name=func.__name__,
                allow_none=allow_none,
            )

            return func(self, value)

        return wrapper

    return decorator



import os
from tempfile import TemporaryDirectory

import pytest

from nxtomo.nxobject.nxobject import NXobject


class test_nx_object:
    """Test API of the NXobject."""

    with pytest.raises(TypeError):
        NXobject(node_name=12)
    with pytest.raises(TypeError):
        NXobject(node_name="test", parent=12)

    nx_object = NXobject(node_name="NXobject")
    with pytest.raises(NotImplementedError):
        nx_object.to_nx_dict(nexus_path_version=1.0)
    assert nx_object.is_root is True

    with pytest.raises(TypeError):
        nx_object.node_name = 12

    with pytest.raises(AttributeError):
        nx_object.test = 12

    class MyNXObject(NXobject):
        def to_nx_dict(
            self,
            nexus_path_version: float | None = None,
            data_path: str | None = None,
        ) -> dict:
            return {
                f"{self.path}/test": "toto",
            }

    my_nx_object = MyNXObject(node_name="NxObject2")

    with TemporaryDirectory() as folder:
        file_path = os.path.join(folder, "my_nexus.nx")
        assert not os.path.exists(file_path)
        my_nx_object.save(
            file_path=file_path, data_path="/object", nexus_path_version=1.0
        )
        assert os.path.exists(file_path)

        with pytest.raises(KeyError):
            my_nx_object.save(
                file_path=file_path,
                data_path="/object",
                nexus_path_version=1.0,
                overwrite=False,
            )

        my_nx_object.save(
            file_path=file_path,
            data_path="/object",
            nexus_path_version=1.0,
            overwrite=True,
        )



"""
Utility helpers for NXobject concatenation.
"""

import numpy
import pint

from nxtomo.nxobject.nxobject import NXobject


def concatenate(
    nx_objects: list[NXobject] | tuple[NXobject, ...], **kwargs
) -> NXobject:
    """
    Concatenate a list of NXobjects.

    :param list | tuple nx_objects: objects to be concatenated. They are expected to be of the same type. Accepts list, or tuple.
    :param kwargs: extra parameters
    :return: concatenated object, of the same type as ``nx_objects``.
    :rtype: :class:`~nxtomo.nxobject.nxobject.NXobject`
    """
    if len(nx_objects) == 0:
        return None
    else:
        if not isinstance(nx_objects[0], NXobject):
            raise TypeError("nx_objects are expected to be instances of NXobject")
        return type(nx_objects[0]).concatenate(nx_objects=nx_objects, **kwargs)


def concatenate_pint_quantities(
    quantities: tuple[pint.Quantity, ...],
) -> pint.Quantity | None:
    """
    Helper function to concatenate pint quantities while ensuring unit consistency.
    """
    if len(quantities) == 0:
        return None
    if len(quantities) == 1:
        return quantities[0]
    for q in quantities:
        if not isinstance(q, pint.Quantity):
            import traceback

            traceback.print_stack(limit=5)
            raise TypeError(
                f"All elements must be pint.Quantity objects. got {type(q)}"
            )
    units = {val.units for val in quantities if isinstance(val, pint.Quantity)}
    if len(units) > 1:
        raise ValueError(f"Inconsistent units {units}")
    unit = units.pop() if units else None
    magnitudes = [val.magnitude for val in quantities]
    concatenated = numpy.concatenate(magnitudes)
    return concatenated * unit if unit else concatenated



"""
Module for handling an `NXtransformations <https://manual.nexusformat.org/classes/base_classes/nxtransformations.html#nxtransformations>`_ group.
"""

import logging
from copy import deepcopy

import h5py

# For unit conversion using pint
import pint
from silx.io.dictdump import nxtodict
from silx.io.utils import open as hdf5_open
from silx.utils.proxy import docstring

from nxtomo.nxobject.nxobject import NXobject
from nxtomo.paths.nxtomo import get_paths as get_nexus_paths
from nxtomo.utils.transformation import GravityTransformation, Transformation
from nxtomo.utils.transformation import get_lr_flip as _get_lr_flip
from nxtomo.utils.transformation import get_ud_flip as _get_ud_flip

_ureg = pint.UnitRegistry()

_logger = logging.getLogger(__name__)

__all__ = ["NXtransformations", "get_lr_flip", "get_ud_flip"]


class NXtransformations(NXobject):
    def __init__(self, node_name: str = "transformations", parent=None) -> None:
        """
        Collection of axis-based translations and rotations to describe a geometry.

        For tomotools the first usage would be to allow users to provide more metadata to tag acquisitions (like "detector has been rotated" by 90 degrees).

        :param node_name: name of the transformations group in the hierarchy.
        :param parent: parent in the NeXus hierarchy.
        """
        super().__init__(node_name, parent)
        self._set_freeze(False)
        self._transformations = dict()
        # dict with axis_name as key and Transformation as value.
        self._set_freeze(True)

    @property
    def transformations(self) -> tuple:
        """
        Return the registered transformations as a tuple.
        """
        return tuple(self._transformations.values())

    @transformations.setter
    def transformations(self, transformations: tuple):
        """
        :param transformations: tuple of Transformation instances
        """
        # check type
        if not isinstance(transformations, (tuple, list)):
            raise TypeError(
                f"transformations is expected to be a dict. {type(transformations)} provided instead"
            )
        for transformation in transformations:
            if not isinstance(transformation, Transformation):
                raise TypeError(
                    f"elements are expected to be instances of {Transformation}. {type(transformation)} provided instead"
                )
        # convert it to a dict for convenience
        self._transformations = {
            transformation.axis_name: transformation
            for transformation in transformations
        }

    def addTransformation(self, *args, **kwargs):
        _logger.warning(
            "addTransformation is deprecated. Please use add_transformation"
        )
        self.add_transformation(*args, **kwargs)

    def add_transformation(
        self, transformation: Transformation, overwrite=False, skip_if_exists=False
    ):
        """
        Add a transformation to the existing ones.

        :param transformation: transformation to be added
        :param overwrite: if a transformation with the same axis_name already exists then overwrite it
        :param skip_if_exists: if a transformation with the same axis_name already exists then keep the existing one
        :raises: KeyError, if a transformation with the same axis_name is already registered
        """
        if skip_if_exists is overwrite is True:
            raise ValueError(
                "both 'skip_if_exists' and 'overwrite' set to True. Undefined behavior"
            )
        if transformation.axis_name in self._transformations:
            if overwrite:
                _logger.info(
                    "A transformation over {transformation.axis_name} is already registered. Will overwrite it"
                )
            elif skip_if_exists:
                _logger.info(
                    "A transformation over {transformation.axis_name} is already registered. Skip add"
                )
                return
            else:
                raise KeyError(
                    f"A transformation over {transformation.axis_name} is already registered. axis_name must be unique"
                )

        self._transformations[transformation.axis_name] = transformation

    def rmTransformation(self, *args, **kwargs):
        _logger.warning("rmTransformation is deprecated. Please use rm_transformation")
        self.rm_transformation(*args, **kwargs)

    def rm_transformation(self, transformation: Transformation):
        """
        Remove the provided transformation from the list of existing transformations.

        :param transformation: transformation to be removed
        """
        self._transformations.pop(transformation.axis_name, None)

    @docstring(NXobject)
    def to_nx_dict(
        self,
        nexus_path_version: float | None = None,
        data_path: str | None = None,
        solve_empty_dependency: bool = False,
    ) -> dict:
        """
        Dump the NXtransformations to a dictionary.

        :param nexus_path_version: Nexus version number.
        :param data_path: Data path where transformations are stored.
        :param solve_empty_dependency: If True, transformations without dependency will be set to depend on gravity.
        """
        if len(self._transformations) == 0:
            # if no transformation, avoid creating the group
            return {}
        nexus_paths = get_nexus_paths(nexus_path_version)
        transformations_nexus_paths = nexus_paths.nx_transformations_paths
        if transformations_nexus_paths is None:
            _logger.info(
                f"no TRANSFORMATIONS provided for version {nexus_path_version}"
            )
            return {}

        transformations = deepcopy(self._transformations)
        # Preprocessing for gravity: set transformations with no dependency to depend on gravity.
        if solve_empty_dependency:
            transformations_needing_gravity = dict(
                filter(
                    lambda pair: pair[1].depends_on in (None, ""),
                    transformations.items(),
                )
            )
            if len(transformations_needing_gravity) > 0:
                gravity = GravityTransformation()
                gravity_name = gravity.axis_name
                if (
                    gravity_name in transformations.keys()
                    and transformations[gravity_name] != gravity
                ):
                    _logger.warning(
                        f"transformations already contains a transformation named '{gravity.axis_name}'. Unable to expand transformation chain (cannot append gravity twice)"
                    )
                else:
                    transformations[gravity_name] = gravity
                # Update transformations needing gravity
                for transformation in transformations_needing_gravity.values():
                    transformation.depends_on = gravity_name

        # Dump each Transformation to a dictionary and adjust the units mapping.
        nx_dict = {}
        for transformation in transformations.values():
            if not isinstance(transformation, Transformation):
                raise TypeError(
                    f"transformations are expected to be instances of {Transformation}. {type(transformation)} provided instead."
                )
            trans_dict = transformation.to_nx_dict(
                transformations_nexus_paths=transformations_nexus_paths,
                data_path=self.path,
            )
            # Adjust unit strings and values according to Nexus conventions:
            # - For rotations, convert to radians and use unit "rad"
            # - For translations, use "m" for meter.
            for key in list(trans_dict.keys()):
                if key.endswith("@units"):
                    unit = trans_dict[key]
                    if unit in ("degree", "degrees") or unit in (
                        "radian",
                        "radians",
                        "rad",
                    ):
                        # Convert the transformation value to degrees regardless of the input unit.
                        q_deg = transformation.transformation_values.to(_ureg.degree)
                        if self.path:
                            value_key = f"{self.path}/{transformation.axis_name}"
                        else:
                            value_key = transformation.axis_name
                        trans_dict[value_key] = q_deg.magnitude
                        trans_dict[key] = "degree"
                    elif unit in ("meter", "meters"):
                        trans_dict[key] = "m"
                    elif unit in ("m/s2", "m/s^2", "m/s**2", "meter / second ** 2"):
                        trans_dict[key] = "m/s^2"
            nx_dict.update(trans_dict)
        nx_dict[f"{self.path}@NX_class"] = "NX_transformations"
        nx_dict[f"{self.path}@units"] = "NX_TRANSFORMATION"
        return nx_dict

    @staticmethod
    def load_from_file(file_path: str, data_path: str, nexus_version: float | None):
        """
        Create an instance of :class:`NXtransformations` and load its value from
        the given file and data path.
        """
        result = NXtransformations()
        return result._load(
            file_path=file_path, data_path=data_path, nexus_version=nexus_version
        )

    def _load(
        self, file_path: str, data_path: str, nexus_version: float | None
    ) -> NXobject:
        """
        Create and load an NXtransformations group from data on disk.
        """
        nexus_paths = get_nexus_paths(nexus_version)
        transformations_nexus_paths = nexus_paths.nx_transformations_paths

        with hdf5_open(file_path) as h5f:
            if data_path == "":
                pass
            elif data_path not in h5f:
                _logger.error(
                    f"No NXtransformations found in {file_path} under {data_path} location."
                )
                return

        transformations_as_nx_dict = nxtodict(file_path, path=data_path)
        # Filter attributes from the dict (as a convention dict contains '@' char)
        transformations_keys = dict(
            filter(lambda a: "@" not in a[0], transformations_as_nx_dict.items())
        )
        for key in transformations_keys:
            transformation = Transformation.from_nx_dict(
                axis_name=key,
                dict_=transformations_as_nx_dict,
                transformations_nexus_paths=transformations_nexus_paths,
            )
            if transformation is None:
                # if failed to load transformation (old version of Nexus?)
                continue
            else:
                self.add_transformation(transformation=transformation)
        return self

    @staticmethod
    @docstring(NXobject)
    def concatenate(nx_objects: tuple, node_name="transformations"):
        res = NXtransformations(node_name=node_name)
        for nx_transformations in nx_objects:
            if not isinstance(nx_transformations, NXtransformations):
                raise TypeError(
                    f"can only concatenate {NXtransformations}. Not {type(nx_transformations)}"
                )
            for transformation in nx_transformations.transformations:
                res.add_transformation(transformation, skip_if_exists=True)
        return res

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, NXtransformations):
            return False
        else:
            # To check equality we filter gravity as it can be provided at the end and as the reference.
            def is_not_gravity(transformation):
                return transformation != GravityTransformation()

            return list(filter(is_not_gravity, self.transformations)) == list(
                filter(is_not_gravity, __value.transformations)
            )

    @staticmethod
    def is_a_valid_group(group: h5py.Group) -> bool:
        """
        Check if the group represents an NXtransformations.
        For now the only condition is to be a group and have NXtransformations as an attribute.
        """
        if not isinstance(group, h5py.Group):
            return False
        return group.attrs.get("NX_class", None) in (
            "NX_transformations",
            "NX_TRANSFORMATIONS",
        )

    def __len__(self):
        return len(self.transformations)


def get_lr_flip(transformations: tuple | NXtransformations) -> tuple:
    """
    Check along all transformations for those matching a left-right flip.

    Return a tuple with all matching transformations.
    """
    if isinstance(transformations, NXtransformations):
        transformations = transformations.transformations
    return _get_lr_flip(transformations)


def get_ud_flip(transformations: tuple | NXtransformations) -> tuple:
    """
    Check along all transformations for those matching an up-down flip.

    Return a tuple with all matching transformations.
    """
    if isinstance(transformations, NXtransformations):
        transformations = transformations.transformations
    return _get_ud_flip(transformations)



import numpy
import pint
import pytest

from nxtomo.nxobject.nxmonitor import NXmonitor
from nxtomo.nxobject.utils.concatenate import concatenate

_ureg = pint.UnitRegistry()


def test_nx_sample():
    """Test creation and saving of an NXmonitor."""
    nx_monitor = NXmonitor()
    # check name
    with pytest.raises(TypeError):
        nx_monitor.data = 12
    with pytest.raises(ValueError):
        nx_monitor.data = numpy.zeros([12, 12])
    nx_monitor.data = tuple()
    nx_monitor.data = numpy.zeros(12)

    assert isinstance(nx_monitor.to_nx_dict(), dict)

    # test concatenate
    nx_monitor_1 = NXmonitor()
    nx_monitor_1.data = numpy.arange(10)
    nx_monitor_2 = NXmonitor()
    nx_monitor_2.data = numpy.arange(10)[::-1] * _ureg.milliampere

    nx_monitor_concat = concatenate([nx_monitor_1, nx_monitor_2])
    assert isinstance(nx_monitor_concat, NXmonitor)
    expected_data = numpy.concatenate(
        [
            nx_monitor_1.data.magnitude,
            nx_monitor_2.data.to(_ureg.ampere).magnitude,
        ]
    )
    numpy.testing.assert_array_equal(
        nx_monitor_concat.data.magnitude,
        expected_data,
    )
    assert str(nx_monitor_concat.data.units) == str(_ureg.ampere)



"""
Module to edit, load, and save following the `NXtomo application definition <https://manual.nexusformat.org/classes/applications/NXtomo.html>`_.
"""

from nxtomo.version import version as __version__  # noqa F401

from .application.nxtomo import NXtomo  # noqa F401



import pint
import pytest

from nxtomo.nxobject.nxsource import NXsource

ureg = pint.get_application_registry()

meter = ureg.meter


def test_nx_source():
    """Test creation and saving of an NXsource."""
    nx_source = NXsource()

    with pytest.raises(TypeError):
        nx_source.name = 12
    nx_source.name = "my source"

    with pytest.raises(AttributeError):
        nx_source.source_name = "test"

    with pytest.raises(ValueError):
        nx_source.type = "toto"
    nx_source.type = "Synchrotron X-ray Source"
    str(nx_source)
    nx_source.type = None
    str(nx_source)

    assert nx_source.probe is None
    nx_source.probe = "neutron"
    assert nx_source.probe.value == "neutron"
    with pytest.raises(ValueError):
        nx_source.probe = 12

    assert nx_source.distance is None

    nx_source.distance = 12.6 * meter
    assert nx_source.distance.magnitude == 12.6
    assert nx_source.distance.units == "meter"

    with pytest.raises(TypeError):
        nx_source.distance = "ddsad"

    assert isinstance(nx_source.to_nx_dict(), dict)

    # Check we can't set undefined attributes
    with pytest.raises(AttributeError):
        nx_source.test = 12

    # Test some concatenation
    nx_source_concatenate = NXsource.concatenate([nx_source, nx_source])
    assert nx_source_concatenate.name == "my source"
    assert nx_source_concatenate.type is None
    assert nx_source_concatenate.probe.value == "neutron"
    assert nx_source_concatenate.distance.magnitude == 12.6
    assert nx_source_concatenate.distance.units == "meter"



"""
Module for handling an `NXdetector <https://manual.nexusformat.org/classes/base_classes/NXdetector.html>`_.
"""

import os
from functools import partial
from operator import is_not

import h5py
import numpy
import pint
from h5py import VirtualSource
from h5py import h5s as h5py_h5s
from silx.io.url import DataUrl
from silx.io.utils import open as hdf5_open
from silx.utils.enum import Enum as _Enum
from silx.utils.proxy import docstring

from nxtomo.io import from_virtual_source_to_data_url
from nxtomo.nxobject.nxobject import NXobject
from nxtomo.nxobject.nxtransformations import (
    NXtransformations,
    get_lr_flip,
    get_ud_flip,
)
from nxtomo.nxobject.utils.decorator import check_dimensionality
from nxtomo.nxobject.utils.ObjectWithPixelSizeMixIn import (
    _ObjectWithPixelSizeMixIn,
    check_quantity_consistency,
)
from nxtomo.paths.nxtomo import get_paths as get_nexus_path
from nxtomo.utils import cast_and_check_array_1D, get_data, get_quantity
from nxtomo.utils.frameappender import FrameAppender
from nxtomo.utils.io import deprecated, ignore_deprecation_warning
from nxtomo.utils.transformation import DetXFlipTransformation, DetYFlipTransformation

try:
    from h5py._hl.vds import VDSmap
except ImportError:
    has_VDSmap = False
else:
    has_VDSmap = True
import logging

import h5py._hl.selections as selection

_logger = logging.getLogger(__name__)


__all__ = ["FOV", "ImageKey", "FieldOfView", "NXdetector", "NXdetectorWithUnit"]

_ureg = pint.UnitRegistry()
_meter = _ureg.meter
_second = _ureg.second


class FOV(_Enum):
    """
    Possible field-of-view values. Use cases are described `here <https://tomotools.gitlab-pages.esrf.fr/tomoscan/language/half_acquisition.html?highlight=half>`_.

    """

    @classmethod
    def from_value(cls, value):
        if isinstance(value, str):
            value = value.lower().title()

        return FOV(value)

    FULL = "Full"
    """We expect to have the full dataset in the field of view."""
    HALF = "Half"
    """We expect to have half of the dataset in the field of view—around 360 degrees. Reconstruction will generate a sinogram of the full dataset."""


FieldOfView = FOV


class ImageKey(_Enum):
    """
    NXdetector `image_key <https://manual.nexusformat.org/classes/base_classes/NXdetector.html#image_key>`_.
    Used to distinguish different frame types.
    """

    ALIGNMENT = -1
    """Used for alignment frames (also known as alignment images)."""
    PROJECTION = 0
    """Projection frames."""
    FLAT_FIELD = 1
    """Flat frames."""
    DARK_FIELD = 2
    """Dark frames."""
    INVALID = 3
    """Invalid frames (ignored during analysis)."""


class NXdetector(NXobject, _ObjectWithPixelSizeMixIn):
    def __init__(
        self,
        node_name="detector",
        parent: NXobject | None = None,
        field_of_view: FOV | None = None,
        expected_dim: tuple | None = None,
    ) -> None:
        """
        Representation of `NeXus NXdetector <https://manual.nexusformat.org/classes/base_classes/NXdetector.html>`_.
        Detector of the acquisition.

        :param node_name: name of the detector in the hierarchy.
        :param parent: parent in the NeXus hierarchy.
        :param field_of_view: field of view of the detector, if known.
        :param expected_dim: expected data dimensions, provided as a tuple of ints to be checked when data is set.
        """
        NXobject.__init__(self, node_name=node_name, parent=parent)
        self._set_freeze(False)
        _ObjectWithPixelSizeMixIn.__init__(self)

        self._expected_dim = expected_dim

        self._data = None
        self.image_key_control = None
        self._transformations = NXtransformations(parent=self)
        self._distance = None  # detector / sample distance
        self.field_of_view = field_of_view
        self._count_time = None
        self.tomo_n = None
        self.group_size = None
        self._roi = None
        self.__master_vds_file = None
        # used to record the virtual dataset set file origin in order to solve relative links
        self._x_rotation_axis_pixel_position: float | None = None
        self._y_rotation_axis_pixel_position: float | None = None
        self._sequence_number: numpy.ndarray[numpy.uint32] | None = None
        """Index of each frame on the acquisition sequence"""

        # as the class is 'freeze' we need to set 'estimated_cor_from_motor' once to make sure the API still exists.
        # the logger filtering avoid to have deprecation logs...
        with ignore_deprecation_warning():
            self.estimated_cor_from_motor = None
        self._set_freeze(True)

    @property
    def data(self) -> numpy.ndarray | tuple | None:
        """
        Detector data (frames).
        Can be None, a NumPy array, or a collection of DataUrl or h5py.VirtualSource objects.
        """
        return self._data

    @data.setter
    def data(self, data: numpy.ndarray | tuple | None):
        if isinstance(data, (tuple, list)) or (
            isinstance(data, numpy.ndarray)
            and data.ndim == 1
            and (self._expected_dim is None or len(self._expected_dim) > 1)
        ):
            for elmt in data:
                if has_VDSmap:
                    if not isinstance(elmt, (DataUrl, VirtualSource, VDSmap)):
                        raise TypeError(
                            f"element of 'data' are expected to be a {len(self._expected_dim)}D numpy array, a list of silx DataUrl or a list of h5py virtualSource. Not {type(elmt)}"
                        )
            data = tuple(data)
        elif isinstance(data, numpy.ndarray):
            if (
                self._expected_dim is not None
                and data is not None
                and data.ndim not in self._expected_dim
            ):
                raise ValueError(
                    f"data is expected to be {len(self._expected_dim)}D not {data.ndim}D"
                )
        elif data is None:
            pass
        else:
            raise TypeError(
                f"data is expected to be an instance of {numpy.ndarray}, None or a list of silx DataUrl or h5py Virtual Source. Not {type(data)}"
            )
        self._data = data

    @property
    def x_rotation_axis_pixel_position(self) -> float:
        """
        Absolute position of the center of rotation in the detector space in X (X being the abscissa).
        Units: pixel.
        """
        return self._x_rotation_axis_pixel_position

    @x_rotation_axis_pixel_position.setter
    def x_rotation_axis_pixel_position(self, value: float | None) -> None:
        if not isinstance(value, (float, type(None))):
            raise TypeError(
                f"x_rotation_axis_pixel_position is expected ot be an instance of {float} or None. Not {type(value)}"
            )
        self._x_rotation_axis_pixel_position = value

    @property
    def y_rotation_axis_pixel_position(self) -> float:
        """
        Absolute position of the center of rotation in the detector space in Y (Y being the ordinate).
        Units: pixel.

        .. warning::

            This field is not handled at the moment by tomotools. Only the X position is handled.
        """
        return self._y_rotation_axis_pixel_position

    @y_rotation_axis_pixel_position.setter
    def y_rotation_axis_pixel_position(self, value: float | None) -> None:
        if not isinstance(value, (float, type(None))):
            raise TypeError(
                f"y_rotation_axis_pixel_position is expected to be an instance of {float} or None. Not {type(value)}"
            )
        self._y_rotation_axis_pixel_position = value

    @deprecated(replacement="set_transformation_from_lr_flipped", since_version="3.0")
    def set_transformation_from_x_flipped(self, flipped: bool | None):
        return self.set_transformation_from_lr_flipped(flipped=flipped)

    def set_transformation_from_lr_flipped(self, flipped: bool | None):
        """Utility function to set transformations from a simple `x_flipped` boolean ((x in the **detector** coordinate system)).
        Used for backward compatibility and convenience.
        """
        # WARNING: moving from two simple boolean to full NXtransformations make the old API very weak. It should be removed
        # soon (but we want to keep the API for at least one release). This is expected to fail except if you stick to {x,y} flips
        if isinstance(flipped, numpy.bool_):
            flipped = bool(flipped)
        if not isinstance(flipped, (bool, type(None))):
            raise TypeError(
                f"x_flipped should be either a (python) boolean or None and is {flipped}, of type {type(flipped)}."
            )
        current_lr_transfs = get_lr_flip(self.transformations)
        for transf in current_lr_transfs:
            self.transformations.rm_transformation(transformation=transf)

        self.transformations.add_transformation(DetYFlipTransformation(flip=flipped))

    @deprecated(replacement="set_transformation_from_lr_flipped", since_version="3.0")
    def set_transformation_from_y_flipped(self, flipped: bool | None):
        return self.set_transformation_from_ud_flipped(flipped=flipped)

    def set_transformation_from_ud_flipped(self, flipped: bool | None):
        """
        Util function to set a detector transformation (y in the **detector** coordinate system).
        """
        # WARNING: moving from two simple boolean to full NXtransformations make the old API very weak. It should be removed
        # soon (but we want to keep the API for at least one release). This is expected to fail except if you stick to {x,y} flips
        if isinstance(flipped, numpy.bool_):
            flipped = bool(flipped)
        if not isinstance(flipped, (bool, type(None))):
            raise TypeError(
                f"y_flipped should be either a (python) boolean or None and is {flipped}, of type {type(flipped)}."
            )

        current_ud_transfs = get_ud_flip(self.transformations)
        for transf in current_ud_transfs:
            self.transformations.rm_transformation(transf)

        self.transformations.add_transformation(DetXFlipTransformation(flip=flipped))

    @property
    def distance(self) -> pint.Quantity | None:
        """
        sample / detector distance as a pint Quantity.
        """
        return self._distance

    @distance.setter
    @check_dimensionality(expected_dimension="[length]")
    def distance(self, value: pint.Quantity | None) -> None:
        self._distance = value

    @property
    def field_of_view(self) -> FieldOfView | None:
        """
        Detector :class:`~nxtomo.nxobject.nxdetector.FieldOfView`.
        """
        return self._field_of_view

    @field_of_view.setter
    def field_of_view(self, field_of_view: FieldOfView | str | None) -> None:
        if field_of_view is not None:
            field_of_view = FOV.from_value(field_of_view)
        self._field_of_view = field_of_view

    @property
    def count_time(self) -> pint.Quantity | None:
        return self._count_time

    @count_time.setter
    @check_dimensionality(expected_dimension="[time]")
    def count_time(self, count_time: None | pint.Quantity):
        self._count_time = count_time

    @property
    @deprecated(
        replacement="x_rotation_axis_pixel_position",
        reason="exists in nexus standard",
        since_version="1.3",
    )
    def estimated_cor_from_motor(self) -> float | None:
        """
        Hint of the center of rotation in pixels read from the motor (when possible).
        """
        return self._x_rotation_axis_pixel_position

    @estimated_cor_from_motor.setter
    @deprecated(
        replacement="x_rotation_axis_pixel_position",
        reason="exists in nexus standard",
        since_version="1.3",
    )
    def estimated_cor_from_motor(self, estimated_cor_from_motor: float | None):
        self._x_rotation_axis_pixel_position = estimated_cor_from_motor

    @property
    def image_key_control(self) -> numpy.ndarray | None:
        """
        :class:`~nxtomo.nxobject.nxdetector.ImageKey` for each frame.
        """
        return self._image_key_control

    @image_key_control.setter
    def image_key_control(
        self,
        control_image_key: list | tuple | numpy.ndarray | None,
    ):
        control_image_key = cast_and_check_array_1D(
            control_image_key, "control_image_key"
        )
        if control_image_key is None:
            self._image_key_control = None
        else:
            # cast all value to instances of ImageKey
            self._image_key_control = numpy.asarray(
                [ImageKey(key) for key in control_image_key]
            )

    @property
    def image_key(self) -> numpy.ndarray | None:
        """
        :class:`~nxtomo.nxobject.nxdetector.ImageKey` for each frame. Replace all
        :class:`~nxtomo.nxobject.nxdetector.ImageKey.ALIGNMENT` values with
        :class:`~nxtomo.nxobject.nxdetector.ImageKey.PROJECTION` to fulfill the NeXus standard.
        """
        if self.image_key_control is None:
            return None
        else:
            control_image_key = self.image_key_control.copy()
            control_image_key[control_image_key == ImageKey.ALIGNMENT] = (
                ImageKey.PROJECTION
            )
            return control_image_key

    @property
    def tomo_n(self) -> int | None:
        """
        Expected number of :class:`~nxtomo.nxobject.nxdetector.ImageKey.PROJECTION` frames.
        """
        return self._tomo_n

    @tomo_n.setter
    def tomo_n(self, tomo_n: int | None):
        self._tomo_n = tomo_n

    @property
    def group_size(self) -> int | None:
        """
        Number of acquisitions for the dataset.
        """
        return self._group_size

    @group_size.setter
    def group_size(self, group_size: int | None):
        self._group_size = group_size

    @property
    def roi(self) -> tuple | None:
        """
        Detector region of interest as (x0, y0, x1, y1).
        """
        return self._roi

    @roi.setter
    def roi(self, roi: tuple | None) -> None:
        if roi is None:
            self._roi = None
        elif not isinstance(roi, (tuple, list, numpy.ndarray)):
            raise TypeError("roi is expected to be None or a tuple")
        elif len(roi) != 4:
            raise ValueError(
                f"roi is expected to contains four elements. Get {len(roi)}"
            )
        else:
            self._roi = tuple(roi)

    @property
    def sequence_number(self) -> numpy.ndarray[numpy.uint32] | None:
        return self._sequence_number

    @sequence_number.setter
    def sequence_number(self, values: numpy.ndarray[numpy.uint32] | None):
        if values is None:
            self._sequence_number = None
        elif (
            isinstance(values, numpy.ndarray)
            and values.dtype == numpy.uint32
            and values.ndim == 1
        ):
            self._sequence_number = values
        else:
            raise TypeError(
                f"'values' is expected to be None or a 1D numpy array of uint32. Got {type(values)}"
            )

    @docstring(NXobject)
    def to_nx_dict(
        self,
        nexus_path_version: float | None = None,
        data_path: str | None = None,
    ) -> dict:
        nexus_paths = get_nexus_path(nexus_path_version)
        nexus_detector_paths = nexus_paths.nx_detector_paths
        x_pixel_size_path = (
            "/".join([self.path, nexus_detector_paths.X_PIXEL_SIZE])
            if nexus_detector_paths.X_PIXEL_SIZE is not None
            else None
        )
        y_pixel_size_path = (
            "/".join([self.path, nexus_detector_paths.Y_PIXEL_SIZE])
            if nexus_detector_paths.Y_PIXEL_SIZE is not None
            else None
        )
        nx_dict = _ObjectWithPixelSizeMixIn.to_nx_dict(
            self,
            x_pixel_size_path=x_pixel_size_path,
            y_pixel_size_path=y_pixel_size_path,
        )

        # image key control
        if self.image_key_control is not None:
            path_img_key = f"{self.path}/{nexus_detector_paths.IMAGE_KEY}"
            nx_dict[path_img_key] = [img_key.value for img_key in self.image_key]
            path_img_key_ctrl = f"{self.path}/{nexus_detector_paths.IMAGE_KEY_CONTROL}"
            nx_dict[path_img_key_ctrl] = [
                img_key.value for img_key in self.image_key_control
            ]
        # distance
        if self.distance is not None:
            path_distance = f"{self.path}/{nexus_detector_paths.DISTANCE}"
            nx_dict[path_distance] = self.distance.magnitude
            nx_dict["@".join([path_distance, "units"])] = f"{self.distance.units:~}"
        # FOV
        if self.field_of_view is not None:
            path_fov = f"{self.path}/{nexus_detector_paths.FOV}"
            nx_dict[path_fov] = self.field_of_view.value
        # count time
        if self.count_time is not None:
            path_count_time = f"{self.path}/{nexus_detector_paths.EXPOSURE_TIME}"
            nx_dict[path_count_time] = self.count_time.magnitude
            nx_dict["@".join([path_count_time, "units"])] = f"{self.count_time.units:~}"
        # tomo n
        if self.tomo_n is not None:
            tomo_n_fov_path = f"{nexus_paths.TOMO_N_SCAN}"
            nx_dict[tomo_n_fov_path] = self.tomo_n
        if self.group_size is not None:
            group_size_path = f"{self.path}/{nexus_paths.GRP_SIZE_ATTR}"
            nx_dict[group_size_path] = self.group_size
        # sequence number
        if (
            self.sequence_number is not None
            and nexus_detector_paths.SEQUENCE_NUMBER is not None
        ):
            sequence_number_path = f"{self.path}/{nexus_detector_paths.SEQUENCE_NUMBER}"
            nx_dict[sequence_number_path] = self.sequence_number

        # x rotation axis position
        if self.x_rotation_axis_pixel_position is not None:
            x_rotation_axis_pixel_position_path = (
                nexus_detector_paths.X_ROTATION_AXIS_PIXEL_POSITION
                or nexus_detector_paths.ESTIMATED_COR_FRM_MOTOR
            )
            if x_rotation_axis_pixel_position_path is not None:
                x_rot_axis_pos_path = (
                    f"{self.path}/{x_rotation_axis_pixel_position_path}"
                )
                nx_dict[x_rot_axis_pos_path] = self.x_rotation_axis_pixel_position
                nx_dict[f"{x_rot_axis_pos_path}@units"] = "pixel"

        # y rotation axis position
        if (
            self.y_rotation_axis_pixel_position is not None
            and nexus_detector_paths.Y_ROTATION_AXIS_PIXEL_POSITION is not None
        ):
            y_rot_axis_pos_path = (
                f"{self.path}/{nexus_detector_paths.Y_ROTATION_AXIS_PIXEL_POSITION}"
            )
            nx_dict[y_rot_axis_pos_path] = self.y_rotation_axis_pixel_position
            nx_dict[f"{y_rot_axis_pos_path}@units"] = "pixel"

        if self.roi is not None:
            path_roi = f"{self.path}/{nexus_detector_paths.ROI}"
            nx_dict[path_roi] = self.roi
            nx_dict["@".join([path_roi, "units"])] = "pixel"

        # export TRANSFORMATIONS
        nx_dict.update(
            self.transformations.to_nx_dict(
                nexus_path_version=nexus_path_version,
                data_path=data_path,
                solve_empty_dependency=True,
            )
        )

        # export detector data
        nx_dict.update(
            self._data_to_nx_dict(
                nexus_path_version=nexus_path_version,
                data_path=data_path,
            )
        )
        return nx_dict

    def _data_to_nx_dict(
        self,
        nexus_path_version: float | None = None,
        data_path: str | None = None,
    ) -> dict:
        nexus_paths = get_nexus_path(nexus_path_version)
        nexus_detector_paths = nexus_paths.nx_detector_paths

        nx_dict = {}
        if self.data is not None:
            # add data
            path_data = f"{self.path}/{nexus_detector_paths.DATA}"
            nx_dict[path_data] = self.data
            nx_dict["@".join([path_data, "interpretation"])] = "image"
            nx_dict["__vds_master_file__"] = self.__master_vds_file
            # add attributes to data
            nx_dict[f"{self.path}@NX_class"] = "NXdetector"
            nx_dict[f"{self.path}@signal"] = nexus_detector_paths.DATA
            nx_dict[f"{self.path}@SILX_style/axis_scale_types"] = [
                "linear",
                "linear",
            ]
        return nx_dict

    def _load(
        self, file_path: str, data_path: str, nexus_version: float, load_data_as: str
    ) -> None:
        possible_as_values = ("as_virtual_source", "as_data_url", "as_numpy_array")
        if load_data_as not in possible_as_values:
            raise ValueError(
                f"load_data_as is expected to be in {possible_as_values} and not {load_data_as}"
            )

        self.__master_vds_file = file_path
        # record the input file if we need to solve virtual dataset path from it

        nexus_paths = get_nexus_path(nexus_version)
        nexus_detector_paths = nexus_paths.nx_detector_paths

        data_dataset_path = f"{data_path}/{nexus_detector_paths.DATA}"

        def vs_file_path_to_real_path(file_path, vs_file_path):
            # get file path as absolute for the NXtomo. Simplify management of the
            # directories
            if os.path.isabs(vs_file_path):
                return vs_file_path
            else:
                return os.path.join(os.path.dirname(file_path), vs_file_path)

        # step 1: load frames
        with hdf5_open(file_path) as h5f:
            if data_dataset_path in h5f:
                dataset = h5f[data_dataset_path]
            else:
                _logger.error(f"unable to find {data_dataset_path} from {file_path}")
                return
            if load_data_as == "as_numpy_array":
                self.data = dataset[()]
            elif load_data_as == "as_data_url":
                if dataset.is_virtual:
                    urls = []
                    for vs_info in dataset.virtual_sources():
                        select_bounds = vs_info.vspace.get_select_bounds()
                        left_bound = select_bounds[0]
                        right_bound = select_bounds[1]
                        # warning: for now step is not managed with virtual
                        # dataset

                        length = right_bound[0] - left_bound[0] + 1
                        # warning: for now step is not managed with virtual
                        # dataset
                        virtual_source = h5py.VirtualSource(
                            vs_file_path_to_real_path(
                                file_path=file_path, vs_file_path=vs_info.file_name
                            ),
                            vs_info.dset_name,
                            vs_info.vspace.shape,
                        )
                        # here we could provide dataset but we won't to
                        # insure file path will be relative.
                        type_code = vs_info.src_space.get_select_type()
                        # check for unlimited selections in case where selection is regular
                        # hyperslab, which is the only allowed case for h5s.UNLIMITED to be
                        # in the selection
                        if (
                            type_code == h5py_h5s.SEL_HYPERSLABS
                            and vs_info.src_space.is_regular_hyperslab()
                        ):
                            (
                                source_start,
                                stride,
                                count,
                                block,
                            ) = vs_info.src_space.get_regular_hyperslab()
                            source_end = source_start[0] + length

                            vs_slice = slice(source_start[0], source_end)
                            sel = selection.select(
                                dataset.shape,
                                vs_slice,
                                dataset=dataset,
                            )
                            virtual_source.sel = sel
                        else:
                            vs_slice = None

                        urls.append(
                            from_virtual_source_to_data_url(
                                virtual_source,
                                vs_slice=vs_slice,
                            )
                        )
                else:
                    urls = [
                        DataUrl(
                            file_path=file_path,
                            data_path=data_dataset_path,
                            scheme="silx",
                        )
                    ]
                self.data = urls
            elif load_data_as == "as_virtual_source":
                if dataset.is_virtual:
                    virtual_sources = []
                    for vs_info in dataset.virtual_sources():
                        u_vs_info = VDSmap(
                            vspace=vs_info.vspace,
                            file_name=vs_file_path_to_real_path(
                                file_path=file_path, vs_file_path=vs_info.file_name
                            ),
                            dset_name=vs_info.dset_name,
                            src_space=vs_info.src_space,
                        )

                        _, vs = FrameAppender._recreate_vs(
                            vs_info=u_vs_info, vds_file=file_path
                        )
                        virtual_sources.append(vs)
                    self.data = virtual_sources
                else:
                    raise ValueError(f"{data_dataset_path} is not virtual")

        # step 2: load metadata
        x_rotation_axis_pixel_position_path = (
            nexus_detector_paths.X_ROTATION_AXIS_PIXEL_POSITION
            or nexus_detector_paths.ESTIMATED_COR_FRM_MOTOR
        )
        if x_rotation_axis_pixel_position_path is not None:
            self.x_rotation_axis_pixel_position = get_data(
                file_path=file_path,
                data_path=f"{data_path}/{x_rotation_axis_pixel_position_path}",
            )

        if nexus_detector_paths.Y_ROTATION_AXIS_PIXEL_POSITION is not None:
            self.y_rotation_axis_pixel_position = get_data(
                file_path=file_path,
                data_path=f"{data_path}/{nexus_detector_paths.Y_ROTATION_AXIS_PIXEL_POSITION}",
            )

        # TODO Henri: create a function without the warning for the backward compatibility
        if nexus_detector_paths.X_FLIPPED is not None:
            self.set_transformation_from_lr_flipped(
                get_data(
                    file_path=file_path,
                    data_path="/".join([data_path, nexus_detector_paths.X_FLIPPED]),
                )
            )
        if nexus_detector_paths.Y_FLIPPED is not None:
            self.set_transformation_from_ud_flipped(
                get_data(
                    file_path=file_path,
                    data_path="/".join([data_path, nexus_detector_paths.Y_FLIPPED]),
                )
            )
        if nexus_detector_paths.NX_TRANSFORMATIONS is not None:
            transformations = self.load_transformations(
                file_path=file_path,
                data_path=data_path,
                nexus_version=nexus_version,
            )
            if transformations is not None:
                transformations.parent = self
                self.transformations = transformations

        try:
            self.distance = get_quantity(
                file_path=file_path,
                data_path="/".join([data_path, nexus_detector_paths.DISTANCE]),
                default_unit=_meter,
            )
        except TypeError as e:
            # in case loaded pixel size doesn't fit the type (case Diamond dataset)
            _logger.warning(f"Fail to load distance. Error is {e}")

        self.field_of_view = get_data(
            file_path=file_path,
            data_path="/".join([data_path, nexus_detector_paths.FOV]),
        )
        self.count_time = get_quantity(
            file_path=file_path,
            data_path="/".join([data_path, nexus_detector_paths.EXPOSURE_TIME]),
            default_unit=_second,
        )
        self.tomo_n = get_data(
            file_path=file_path,
            data_path="/".join([data_path, nexus_paths.TOMO_N_SCAN]),
        )
        self.group_size = get_data(
            file_path=file_path,
            data_path="/".join([data_path, nexus_paths.GRP_SIZE_ATTR]),
        )
        self.image_key_control = get_data(
            file_path=file_path,
            data_path="/".join([data_path, nexus_detector_paths.IMAGE_KEY_CONTROL]),
        )
        if self.image_key_control is None:
            # in the case image_key_control doesn't exists (dimaond dataset use case)
            self.image_key_control = get_data(
                file_path=file_path,
                data_path="/".join([data_path, nexus_detector_paths.IMAGE_KEY]),
            )
        if nexus_detector_paths.SEQUENCE_NUMBER is not None:
            self.sequence_number = get_data(
                file_path=file_path,
                data_path="/".join([data_path, nexus_detector_paths.SEQUENCE_NUMBER]),
            )
        roi = get_data(
            file_path=file_path,
            data_path="/".join([data_path, nexus_detector_paths.ROI]),
        )
        if roi is not None:
            self.roi = roi

        _ObjectWithPixelSizeMixIn._load(
            self,
            file_path=file_path,
            x_pixel_size_path=(
                "/".join([data_path, nexus_detector_paths.X_PIXEL_SIZE])
                if nexus_detector_paths.X_PIXEL_SIZE
                else None
            ),
            y_pixel_size_path=(
                "/".join([data_path, nexus_detector_paths.Y_PIXEL_SIZE])
                if nexus_detector_paths.Y_PIXEL_SIZE
                else None
            ),
        )

    @staticmethod
    def load_transformations(
        file_path: str, data_path: str, nexus_version
    ) -> NXtransformations | None:
        """
        Transformations are not stored at a fixed position: try to load them from
        the default location ('transformations'); otherwise, browse all HDF5 groups
        to retrieve an NXtransformations group.
        """
        nexus_paths = get_nexus_path(nexus_version)
        nexus_detector_paths = nexus_paths.nx_detector_paths

        with hdf5_open(file_path) as h5f:
            if data_path not in h5f:
                return None

            detector_grp = h5f[data_path]
            # filter valid groups (fitting NXtransformations definition)
            valid_data_paths = dict(
                filter(
                    lambda item: NXtransformations.is_a_valid_group(item[1]),
                    detector_grp.items(),
                )
            )

        if len(valid_data_paths) == 0:
            return None
        elif len(valid_data_paths) > 1:
            issue = "more than one NXtransformations group found"
            if nexus_detector_paths.NX_TRANSFORMATIONS in valid_data_paths:
                _logger.warning(
                    f"{issue}. Will pick the default path as there ({nexus_detector_paths.NX_TRANSFORMATIONS})"
                )
                return NXtransformations.load_from_file(
                    file_path=file_path,
                    data_path="/".join(
                        [data_path, nexus_detector_paths.NX_TRANSFORMATIONS]
                    ),
                    nexus_version=nexus_version,
                )
            raise ValueError(f"{issue} - ({valid_data_paths}). Unable to handle it")
        else:
            return NXtransformations.load_from_file(
                file_path=file_path,
                data_path="/".join([data_path, list(valid_data_paths.keys())[0]]),
                nexus_version=nexus_version,
            )

    @staticmethod
    def _concatenate_except_data(nx_detector, nx_objects: tuple):
        image_key_ctrl = [
            nx_obj.image_key_control
            for nx_obj in nx_objects
            if nx_obj.image_key_control is not None
        ]
        if len(image_key_ctrl) > 0:
            nx_detector.image_key_control = numpy.concatenate(image_key_ctrl)

        _ObjectWithPixelSizeMixIn.concatenate(
            output_nx_object=nx_detector, nx_objects=nx_objects
        )
        # note: image_key is deduced from image_key_control
        nx_detector.x_pixel_size = nx_objects[0].x_pixel_size
        nx_detector.roi = nx_objects[0].roi
        nx_detector.y_pixel_size = nx_objects[0].y_pixel_size
        nx_detector.x_rotation_axis_pixel_position = nx_objects[
            0
        ].x_rotation_axis_pixel_position
        nx_detector.y_rotation_axis_pixel_position = nx_objects[
            0
        ].y_rotation_axis_pixel_position
        nx_detector.roi = nx_objects[0].roi
        nx_detector.distance = nx_objects[0].distance
        nx_detector.field_of_view = nx_objects[0].field_of_view
        nx_detector.transformations = nx_objects[0].transformations
        for nx_obj in nx_objects[1:]:
            if nx_detector.transformations != nx_obj.transformations:
                _logger.warning(
                    f"found different NXTransformations. ({nx_detector.transformations.to_nx_dict()} vs {nx_obj.transformations.to_nx_dict()}). Pick the first one"
                )

            check_quantity_consistency(
                reference=nx_detector.distance,
                candidate=nx_obj.distance,
                label="detector distance",
                logger=_logger,
            )
            if (
                nx_detector.field_of_view
                and nx_detector.field_of_view != nx_obj.field_of_view
            ):
                _logger.warning(
                    f"found different field_of_view value. ({nx_detector.field_of_view} vs {nx_obj.field_of_view}). Pick the first one"
                )
            if nx_detector.roi != nx_obj.roi:
                _logger.warning(
                    f"found different detector roi value. ({nx_detector.roi} vs {nx_obj.roi}). Pick the first one"
                )

    @staticmethod
    @docstring(NXobject)
    def concatenate(nx_objects: tuple, node_name="detector"):
        # filter None obj
        nx_objects = tuple(filter(partial(is_not, None), nx_objects))
        if len(nx_objects) == 0:
            return None
        # warning: later we make the assumption that nx_objects contains at least one element
        for nx_obj in nx_objects:
            if not isinstance(nx_obj, NXdetector):
                raise TypeError("Cannot concatenate non NXinstrument object")

        nx_detector = NXdetector(node_name=node_name)
        NXdetector._concatenate_except_data(
            nx_objects=nx_objects, nx_detector=nx_detector
        )

        # now handle data on it's own
        detector_data = [
            nx_obj.data for nx_obj in nx_objects if nx_obj.data is not None
        ]
        if len(detector_data) > 0:
            if isinstance(detector_data[0], numpy.ndarray):
                # store_as = "as_numpy_array"
                expected = numpy.ndarray
            elif isinstance(detector_data[0], (list, tuple)):
                if isinstance(detector_data[0][0], h5py.VirtualSource):
                    # store_as = "as_virtual_source"
                    expected = h5py.VirtualSource
                elif isinstance(detector_data[0][0], DataUrl):
                    # store_as = "as_data_url"
                    expected = DataUrl
                else:
                    raise TypeError(
                        f"detector data is expected to be a numpy array or a h5py.VirtualSource or a numpy array. {type(detector_data[0][0])} is not handled."
                    )
            else:
                raise TypeError(
                    f"detector data is expected to be a numpy array or a h5py.VirtualSource or a numpy array. {type(detector_data[0])} is not handled."
                )

            for data in detector_data:
                if expected in (DataUrl, h5py.VirtualSource):
                    # for DataUrl and VirtualSource check type of the element
                    cond = isinstance(data[0], expected)
                else:
                    cond = isinstance(data, expected)
                if not cond:
                    raise TypeError(
                        f"Incoherent data type cross detector data ({type(data)} when {expected} expected)"
                    )

            if expected in (DataUrl, h5py.VirtualSource):
                new_data = []
                [new_data.extend(data) for data in detector_data]
            else:
                new_data = numpy.concatenate(detector_data)
            nx_detector.data = new_data

        return nx_detector

    @property
    def transformations(self):
        """
        `NXtransformation <https://manual.nexusformat.org/classes/base_classes/NXtransformations.html>`_
        objects describing detector flips or manual rotations.
        """
        return self._transformations

    @transformations.setter
    def transformations(self, transformations: NXtransformations) -> None:
        self._transformations = transformations


class NXdetectorWithUnit(NXdetector):
    def __init__(
        self,
        default_unit: pint.Unit,
        node_name="detector",
        parent=None,
        field_of_view=None,
        expected_dim: tuple | None = None,
    ) -> None:
        self._default_unit = default_unit
        super().__init__(node_name, parent, field_of_view, expected_dim)
        self._data = None

    @property
    def data(self) -> pint.Quantity | None:
        """
        Detector data.
        Can be a NumPy array, list of DataUrl objects, VirtualSource instances, or a pint.Quantity.
        """
        return self._data

    @data.setter
    def data(self, data: pint.Quantity | numpy.ndarray | tuple | None):
        if isinstance(data, pint.Quantity):
            # Ensure that the magnitude is a NumPy array
            if not isinstance(data.magnitude, numpy.ndarray):
                raise TypeError(
                    "pint.Quantity must have a NumPy array as its magnitude."
                )
            if (
                self._expected_dim is not None
                and data.magnitude.ndim not in self._expected_dim
            ):
                raise ValueError(
                    f"data is expected to be {len(self._expected_dim)}D, not {data.magnitude.ndim}D"
                )
        elif isinstance(data, (tuple, list)) or (
            isinstance(data, numpy.ndarray)
            and data.ndim == 1
            and (self._expected_dim is None or len(self._expected_dim) > 1)
        ):
            for elmt in data:
                if has_VDSmap:
                    if not isinstance(elmt, (DataUrl, VirtualSource, VDSmap)):
                        raise TypeError(
                            f"Elements of 'data' are expected to be a {len(self._expected_dim)}D numpy array, "
                            f"a list of silx DataUrl, or a list of h5py VirtualSource. Not {type(elmt)}."
                        )
            data = tuple(data)
        elif isinstance(data, numpy.ndarray):
            if self._expected_dim is not None and data.ndim not in self._expected_dim:
                raise ValueError(
                    f"data is expected to be {len(self._expected_dim)}D, not {data.ndim}D"
                )
        elif data is None:
            pass
        else:
            raise TypeError(
                f"data is expected to be a pint.Quantity, numpy.ndarray, None, or a list of silx DataUrl/h5py Virtual Source. "
                f"Not {type(data)}."
            )

        self._data = data

    def _data_to_nx_dict(
        self,
        nexus_path_version: float | None = None,
        data_path: str | None = None,
    ) -> dict:
        nexus_paths = get_nexus_path(nexus_path_version)
        nexus_detector_paths = nexus_paths.nx_detector_paths

        nx_dict = {}
        if self.data is not None:
            # add data
            path_data = f"{self.path}/{nexus_detector_paths.DATA}"
            nx_dict[path_data] = self.data
            nx_dict["@".join([path_data, "interpretation"])] = "image"
            # add attributes to data
            nx_dict[f"{self.path}@NX_class"] = "NXdetector"
            nx_dict[f"{self.path}@signal"] = nexus_detector_paths.DATA
            nx_dict[f"{self.path}@SILX_style/axis_scale_types"] = [
                "linear",
                "linear",
            ]
        return nx_dict

    @staticmethod
    @docstring(NXobject)
    def concatenate(
        nx_objects: tuple, default_unit, expected_dim, node_name="detector"
    ):
        # Filter out None objects
        nx_objects = tuple(filter(partial(is_not, None), nx_objects))
        if len(nx_objects) == 0:
            return None

        # Ensure all objects are NXdetector instances
        for nx_obj in nx_objects:
            if not isinstance(nx_obj, NXdetector):
                raise TypeError("Cannot concatenate non-NXdetector object")

        # Create new detector instance
        nx_detector = NXdetectorWithUnit(
            node_name=node_name, default_unit=default_unit, expected_dim=expected_dim
        )
        NXdetector._concatenate_except_data(
            nx_objects=nx_objects, nx_detector=nx_detector
        )

        # Handle data concatenation
        detector_data = [
            nx_obj.data for nx_obj in nx_objects if nx_obj.data is not None
        ]

        # Ensure unit consistency
        detector_units = set(
            nx_obj.data.units
            for nx_obj in nx_objects
            if nx_obj.data is not None and isinstance(nx_obj.data, pint.Quantity)
        )

        if len(detector_units) > 1:
            raise ValueError("More than one unit found. Unable to build the detector.")

        # If no data is present, return early
        if not detector_data or len(detector_data) == 0:
            return nx_detector

        expected_unit = list(detector_units)[0] if detector_units else default_unit

        # Check data type and expected structure
        first_data = detector_data[0]

        if isinstance(first_data, pint.Quantity):
            expected = pint.Quantity
        elif isinstance(first_data, numpy.ndarray):
            expected = numpy.ndarray
        elif isinstance(first_data, (list, tuple)):
            if isinstance(first_data[0], h5py.VirtualSource):
                expected = h5py.VirtualSource
            elif isinstance(first_data[0], DataUrl):
                expected = DataUrl
            else:
                raise TypeError(
                    f"Detector data must be a numpy array, h5py.VirtualSource, or DataUrl. "
                    f"Found {type(first_data[0])}."
                )
        else:
            raise TypeError(
                f"Detector data must be a numpy array, h5py.VirtualSource, or DataUrl. "
                f"Found {type(first_data)}."
            )

        # Validate all data entries
        for data in detector_data:
            if expected in (DataUrl, h5py.VirtualSource):
                cond = isinstance(data[0], expected)
            else:
                cond = isinstance(data, expected) or isinstance(data, pint.Quantity)

            if not cond:
                raise TypeError(
                    f"Incoherent data type across detector data: {type(data)} when {expected} expected."
                )

        # Perform concatenation with proper unit handling
        if expected in (DataUrl, h5py.VirtualSource):
            new_data = []
            for data in detector_data:
                new_data.extend(data)
        else:
            # If the data is pint.Quantity, concatenate magnitudes and reapply the unit
            if isinstance(first_data, pint.Quantity):
                new_data = (
                    numpy.concatenate([data.magnitude for data in detector_data])
                    * expected_unit
                )
            else:
                new_data = numpy.concatenate(detector_data)

        # Assign data correctly
        nx_detector.data = new_data

        return nx_detector



"""
Module for handling an `NXsource <https://manual.nexusformat.org/classes/base_classes/NXsource.html>`_.
"""

import logging
from functools import partial
from operator import is_not

import numpy
import pint
from silx.utils.enum import Enum as _Enum
from silx.utils.proxy import docstring

from nxtomo.nxobject.nxobject import NXobject
from nxtomo.nxobject.utils.decorator import check_dimensionality
from nxtomo.paths.nxtomo import get_paths as get_nexus_paths
from nxtomo.utils import get_data, get_quantity

_logger = logging.getLogger(__name__)

__all__ = ["SourceType", "ProbeType", "NXsource", "DefaultESRFSource"]

_ureg = pint.UnitRegistry()
_meter = _ureg.meter


class SourceType(_Enum):
    """
    Source types like "Synchrotron X-ray Source" or "Free-Electron Laser".
    """

    SPALLATION_NEUTRON = "Spallation Neutron Source"
    PULSED_REACTOR_NEUTRON_SOURCE = "Pulsed Reactor Neutron Source"
    REACTOR_NEUTRON_SOURCE = "Reactor Neutron Source"
    SYNCHROTRON_X_RAY_SOURCE = "Synchrotron X-ray Source"
    PULSED_MUON_SOURCE = "Pulsed Muon Source"
    ROTATING_ANODE_X_RAY = "Rotating Anode X-ray"
    FIXED_TUBE_X_RAY = "Fixed Tube X-ray"
    UV_LASER = "UV Laser"
    FREE_ELECTRON_LASER = "Free-Electron Laser"
    OPTICAL_LASER = "Optical Laser"
    ION_SOURCE = "Ion Source"
    UV_PLASMA_SOURCE = "UV Plasma Source"
    METAL_JET_X_RAY = "Metal Jet X-ray"


class ProbeType(_Enum):
    """
    Probe types like "x-ray" or "neutron".
    """

    NEUTRON = "neutron"
    X_RAY = "x-ray"
    MUON = "muon"
    ELECTRON = "electron"
    ULTRAVIOLET = "ultraviolet"
    VISIBLE_LIGHT = "visible light"
    POSITRON = "positron"
    PROTON = "proton"


class NXsource(NXobject):
    """Information regarding the X-ray storage ring or facility."""

    def __init__(
        self,
        node_name="source",
        parent=None,
        source_name=None,
        source_type=None,
        probe=None,
    ):
        """
        Representation of `NeXus NXsource <https://manual.nexusformat.org/classes/base_classes/NXsource.html>`_.
        The neutron or X-ray storage ring or facility.

        :param node_name: name of the source in the hierarchy.
        :param parent: parent in the NeXus hierarchy.
        :param source_name: name of the source.
        :param source_type: source type.
        :param probe: probe.
        """
        super().__init__(node_name=node_name, parent=parent)
        self._set_freeze(False)
        self._name = source_name
        self._type = source_type
        self._probe = probe
        self._distance = None
        """Source / sample distance."""
        self._set_freeze(True)

    @property
    def name(self) -> None | str:
        """
        Source name.
        """
        return self._name

    @name.setter
    def name(self, source_name: str | None):
        if isinstance(source_name, numpy.ndarray):
            # handle Diamond Dataset
            source_name = source_name.tostring()
            if hasattr(source_name, "decode"):
                source_name = source_name.decode()
        if not isinstance(source_name, (str, type(None))):
            raise TypeError(
                f"source_name is expected to be None or a str not {type(source_name)}"
            )
        self._name = source_name

    @property
    def type(self) -> SourceType | None:
        """
        Source type as :class:`~nxtomo.nxobject.nxsource.SourceType`.
        """
        return self._type

    @type.setter
    def type(self, type_: None | str | SourceType):
        if type_ is None:
            self._type = None
        else:
            type_ = SourceType(type_)
            self._type = type_

    @property
    def probe(self) -> ProbeType | None:
        """
        Probe as :class:`~nxtomo.nxobject.nxsource.ProbeType`.
        """
        return self._probe

    @probe.setter
    def probe(self, probe: None | str | ProbeType):
        if probe is None:
            self._probe = None
        else:
            self._probe = ProbeType(probe)

    @property
    def distance(self) -> pint.Quantity | None:
        return self._distance

    @distance.setter
    @check_dimensionality("[length]")
    def distance(self, value) -> pint.Quantity | None:
        self._distance = value

    def __str__(self):
        return f"{super().__str__}, (source name: {self.name}, source type: {self.type}, source probe: {self.probe})"

    @docstring(NXobject)
    def to_nx_dict(
        self,
        nexus_path_version: float | None = None,
        data_path: str | None = None,
    ) -> dict:
        nexus_paths = get_nexus_paths(nexus_path_version)
        nexus_source_paths = nexus_paths.nx_source_paths
        nx_dict = {}

        # warning: source is integrated only since 1.1 version of the nexus path
        if self.name is not None and nexus_paths.SOURCE_NAME is not None:
            path_name = f"{self.path}/{nexus_source_paths.NAME}"
            nx_dict[path_name] = self.name
        if self.type is not None and nexus_paths.SOURCE_TYPE is not None:
            path_type = f"{self.path}/{nexus_source_paths.TYPE}"
            nx_dict[path_type] = self.type.value
        if self.probe is not None and nexus_paths.SOURCE_PROBE is not None:
            path_probe = f"{self.path}/{nexus_source_paths.PROBE}"
            nx_dict[path_probe] = self.probe.value
        if self.distance is not None and nexus_source_paths.DISTANCE is not None:
            path_source = f"{self.path}/{nexus_source_paths.DISTANCE}"
            nx_dict[path_source] = self.distance.magnitude
            nx_dict["@".join([path_source, "units"])] = f"{self.distance.units:~}"

        # complete the nexus metadata if not empty
        if nx_dict != {}:
            nx_dict[f"{self.path}@NX_class"] = "NXsource"

        return nx_dict

    def _load(self, file_path: str, data_path: str, nexus_version: float) -> None:
        nexus_paths = get_nexus_paths(nexus_version)
        nexus_source_paths = nexus_paths.nx_source_paths
        self.name = get_data(
            file_path=file_path,
            data_path="/".join([data_path, nexus_source_paths.NAME]),
        )
        try:
            self.type = get_data(
                file_path=file_path,
                data_path="/".join([data_path, nexus_source_paths.TYPE]),
            )
        except ValueError as e:
            _logger.warning(f"Fail to load source type. Error is {e}")
        try:
            self.probe = get_data(
                file_path=file_path,
                data_path="/".join([data_path, nexus_source_paths.PROBE]),
            )
        except ValueError as e:
            _logger.warning(f"Fail to load probe. Error is {e}")

        try:
            self.distance = get_quantity(
                file_path=file_path,
                data_path="/".join([data_path, nexus_source_paths.DISTANCE]),
                default_unit=_meter,
            )
        except TypeError as e:
            # in case loaded pixel size doesn't fit the type (case Diamond dataset)
            _logger.warning(f"Fail to load distance. Error is {e}")

    @staticmethod
    @docstring(NXobject)
    def concatenate(nx_objects: tuple, node_name="source"):
        # filter None obj
        nx_objects = tuple(filter(partial(is_not, None), nx_objects))
        if len(nx_objects) == 0:
            return None
        # warning: later we make the assumption that nx_objects contains at least one element
        for nx_obj in nx_objects:
            if not isinstance(nx_obj, NXsource):
                raise TypeError("Cannot concatenate non NXsource object")

        nx_souce = NXsource(node_name=node_name)
        nx_souce.name = nx_objects[0].name
        _logger.info(f"Take the first source name {nx_objects[0].name}")
        nx_souce.type = nx_objects[0].type
        _logger.info(f"Take the first source type {nx_objects[0].type}")
        nx_souce.probe = nx_objects[0].probe
        _logger.info(f"Take the first source probe {nx_objects[0].probe}")
        nx_souce.distance = nx_objects[0].distance
        _logger.info(f"Take the first source distance {nx_objects[0].distance}")
        return nx_souce


class DefaultESRFSource(NXsource):
    """
    ESRF source.
    """

    def __init__(self, node_name="source", parent=None) -> None:
        super().__init__(
            node_name=node_name,
            parent=parent,
            source_name="ESRF",
            source_type=SourceType.SYNCHROTRON_X_RAY_SOURCE,
            probe=ProbeType.X_RAY,
        )



"""
Shared mixins for NXobject pixel size handling.
"""

import logging

import numpy
import pint
from silx.utils.proxy import docstring

from nxtomo.nxobject.nxobject import NXobject
from nxtomo.utils import get_quantity

_logger = logging.getLogger(__name__)
_ureg = pint.get_application_registry()


class _ObjectWithPixelSizeMixIn:
    """
    Class to be shared by NXobject classes that can define a pixel size.
    """

    def __init__(self):
        self._x_pixel_size: pint.Quantity | None = None
        # x 'sample' detector size
        self._y_pixel_size: pint.Quantity | None = None
        # y 'sample' detector size

    @property
    def x_pixel_size(self) -> pint.Quantity | None:
        """
        X pixel size stored as a quantity with units (default unit is SI).
        Known as the "X sample pixel size" in some applications.
        """
        return self._x_pixel_size

    @x_pixel_size.setter
    def x_pixel_size(self, x_pixel_size: pint.Quantity | None) -> None:
        if not isinstance(x_pixel_size, (type(None), pint.Quantity)):
            raise TypeError(
                f"x_pixel_size is expected ot be an instance of {pint.Quantity} or None. Not {type(x_pixel_size)}"
            )
        self._x_pixel_size = x_pixel_size

    @property
    def y_pixel_size(self) -> pint.Quantity | None:
        """
        Y pixel size stored as a quantity with units (default unit is SI).
        Known as the "Y sample pixel size" in some applications.
        """
        return self._y_pixel_size

    @y_pixel_size.setter
    def y_pixel_size(self, y_pixel_size: pint.Quantity | None) -> None:
        if not isinstance(y_pixel_size, (type(None), pint.Quantity)):
            raise TypeError(
                f"y_pixel_size is expected ot be an instance of {pint.Quantity} or None. Not {type(y_pixel_size)}"
            )
        self._y_pixel_size = y_pixel_size

    @docstring(NXobject)
    def to_nx_dict(
        self,
        x_pixel_size_path: str | None,
        y_pixel_size_path: str | None,
    ) -> dict:
        assert isinstance(self, NXobject)

        nx_dict = {}
        # x 'sample' pixel
        if x_pixel_size_path is not None and self.x_pixel_size is not None:
            path_x_pixel_size = x_pixel_size_path  # pylint: disable=E1101
            nx_dict[path_x_pixel_size] = self.x_pixel_size.magnitude
            nx_dict["@".join([path_x_pixel_size, "units"])] = (
                f"{self.x_pixel_size.units:~}"
            )
        # y 'sample' pixel
        if y_pixel_size_path is not None and self.y_pixel_size is not None:
            path_y_pixel_size = y_pixel_size_path
            nx_dict[path_y_pixel_size] = self.y_pixel_size.magnitude
            nx_dict["@".join([path_y_pixel_size, "units"])] = (
                f"{self.y_pixel_size.units:~}"
            )
        return nx_dict

    def _load(
        self,
        file_path: str,
        x_pixel_size_path: str | None,
        y_pixel_size_path: str | None,
    ) -> None:
        # nexus_paths = get_nexus_path(nexus_version)
        # nexus_detector_paths = nexus_paths.nx_detector_paths
        if x_pixel_size_path is not None:
            try:
                self.x_pixel_size = get_quantity(
                    file_path=file_path,
                    data_path=x_pixel_size_path,
                    default_unit=_ureg.meter,
                )
            except TypeError as e:
                # in case loaded pixel size doesn't fit the type (case Diamond dataset)
                _logger.warning(f"Fail to load x pixel size. Error is {e}")
        if y_pixel_size_path is not None:
            try:
                self.y_pixel_size = get_quantity(
                    file_path=file_path,
                    data_path=y_pixel_size_path,  # "/".join([data_path, nexus_detector_paths.Y_PIXEL_SIZE]),
                    default_unit=_ureg.meter,
                )
            except TypeError as e:
                # in case loaded pixel size doesn't fit the type (case Diamond dataset)
                _logger.warning(f"Fail to load y pixel size. Error is {e}")

    @staticmethod
    def concatenate(output_nx_object, nx_objects: tuple):
        """
        Update `output_nx_object` (expected to be the inheriting class) with the pixel size from `nx_objects`.
        """
        if not isinstance(output_nx_object, _ObjectWithPixelSizeMixIn):
            raise TypeError
        if not isinstance(nx_objects[0], _ObjectWithPixelSizeMixIn):
            raise TypeError
        output_nx_object.x_pixel_size = nx_objects[0].x_pixel_size
        output_nx_object.y_pixel_size = nx_objects[0].y_pixel_size
        for nx_obj in nx_objects[1:]:
            check_quantity_consistency(
                reference=output_nx_object.x_pixel_size,
                candidate=nx_obj.x_pixel_size,
                label="x pixel size",
                logger=_logger,
            )
            check_quantity_consistency(
                reference=output_nx_object.y_pixel_size,
                candidate=nx_obj.y_pixel_size,
                label="y pixel size",
                logger=_logger,
            )


def check_quantity_consistency(
    reference: pint.Quantity | None,
    candidate: pint.Quantity | None,
    label: str,
    logger: logging.Logger,
) -> None:
    """
    Compare two pint quantities and warn (or raise) when they differ.

    :param reference: quantity selected for the output object.
    :param candidate: quantity originating from the object being merged.
    :param label: human-readable label used in warning messages.
    :param logger: logger used to emit warnings.
    """
    if reference is None or candidate is None:
        return

    candidate_in_reference_unit = candidate.to(reference.units)

    if not numpy.isclose(reference.magnitude, candidate_in_reference_unit.magnitude):
        logger.warning(
            "found different %s value. (%s vs %s). Pick the first one",
            label,
            reference,
            candidate,
        )



"""I/O utilities."""

import contextlib
import functools
import logging
import traceback
from contextlib import contextmanager

import h5py

try:
    import hdf5plugin  # noqa F401
except ImportError:
    pass
from silx.io.url import DataUrl
from silx.io.utils import open as hdf5_open

__all__ = ["EntryReader", "DatasetReader", "deprecated_warning", "deprecated"]


class _BaseReader(contextlib.AbstractContextManager):
    def __init__(self, url: DataUrl):
        if not isinstance(url, DataUrl):
            raise TypeError(f"url should be an instance of DataUrl. Not {type(url)}")
        if url.scheme() not in ("silx", "h5py"):
            raise ValueError("Valid scheme are silx and h5py")
        if url.data_slice() is not None:
            raise ValueError(
                "Data slices are not managed. Data path should "
                "point to a bliss node (h5py.Group)"
            )
        self._url = url
        self._file_handler = None

    def __exit__(self, *exc):
        return self._file_handler.close()


class EntryReader(_BaseReader):
    """Context manager used to read a BLISS node."""

    def __enter__(self):
        self._file_handler = hdf5_open(filename=self._url.file_path())
        if self._url.data_path() == "":
            entry = self._file_handler
        elif self._url.data_path() not in self._file_handler:
            raise KeyError(
                f"data path '{self._url.data_path()}' doesn't exists from '{self._url.file_path()}'"
            )
        else:
            entry = self._file_handler[self._url.data_path()]
        if not isinstance(entry, h5py.Group):
            raise ValueError("Data path should point to a bliss node (h5py.Group)")
        return entry


class DatasetReader(_BaseReader):
    """Context manager used to read a BLISS node."""

    def __enter__(self):
        self._file_handler = hdf5_open(filename=self._url.file_path())
        entry = self._file_handler[self._url.data_path()]
        if not isinstance(entry, h5py.Dataset):
            raise ValueError(
                f"Data path ({self._url.path()}) should point to a dataset (h5py.Dataset)"
            )
        return entry


depreclog = logging.getLogger("nxtomo.DEPRECATION")

deprecache = set([])


def deprecated_warning(
    type_,
    name,
    reason=None,
    replacement=None,
    since_version=None,
    only_once=True,
    skip_backtrace_count=0,
):
    """
    Log a deprecation warning.

    :param type_: Nature of the object to be deprecated,
        e.g. "Module", "Function", "Class".
    :param name: Object name.
    :param reason: Reason for deprecating this object (e.g. "feature no longer provided").
    :param replacement: Name of the replacement function (when the deprecation renames the function).
    :param since_version: First *silx* version for which the function was deprecated (e.g. "0.5.0").
    :param only_once: If True, the deprecation warning is generated only once for each call site. Default is True.
    :param skip_backtrace_count: Number of trailing stack frames to ignore when logging the backtrace.
    """
    if not depreclog.isEnabledFor(logging.WARNING):
        # Avoid computation when it is not logged
        return

    msg = "%s %s is deprecated"
    if since_version is not None:
        msg += " since silx version %s" % since_version
    msg += "."
    if reason is not None:
        msg += " Reason: %s." % reason
    if replacement is not None:
        msg += " Use '%s' instead." % replacement
    msg += "\n%s"
    limit = 2 + skip_backtrace_count
    backtrace = "".join(traceback.format_stack(limit=limit)[0])
    backtrace = backtrace.rstrip()
    if only_once:
        data = (msg, type_, name, backtrace)
        if data in deprecache:
            return
        else:
            deprecache.add(data)
    depreclog.warning(msg, type_, name, backtrace)


def deprecated(
    func=None,
    reason=None,
    replacement=None,
    since_version=None,
    only_once=True,
    skip_backtrace_count=1,
):
    """
    Decorator that deprecates the use of a function.

    :param str reason: Reason for deprecating this function (e.g. "feature no longer provided").
    :param str replacement: Name of the replacement function (when the deprecation renames the function).
    :param str since_version: First *silx* version for which the function was deprecated (e.g. "0.5.0").
    :param bool only_once: If True, the deprecation warning is generated only once. Default is True.
    :param int skip_backtrace_count: Number of trailing stack frames to ignore when logging the backtrace.
    """

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            deprecated_warning(
                type_="Function",
                name=func.__name__,
                reason=reason,
                replacement=replacement,
                since_version=since_version,
                only_once=only_once,
                skip_backtrace_count=skip_backtrace_count,
            )
            return func(*args, **kwargs)

        return wrapper

    if func is not None:
        return decorator(func)
    return decorator


@contextmanager
def ignore_deprecation_warning():
    """Filter logs from 'nxtomo.DEPRECATION'."""

    def filter(record):
        return record.name != depreclog.name

    depreclog.addFilter(filter)
    yield
    depreclog.removeFilter(filter)



import numpy
import pint
import pytest

from nxtomo.nxobject.nxsample import NXsample

ureg = pint.get_application_registry()
degree = ureg.degree
meter = ureg.meter


def test_nx_sample():
    """Test creation and saving of an NXsample."""
    nx_sample = NXsample()
    # check name
    with pytest.raises(TypeError):
        nx_sample.name = 12
    nx_sample.name = "my sample"

    # check rotation angle
    with pytest.raises(TypeError):
        nx_sample.rotation_angle = 56
    nx_sample.rotation_angle = numpy.linspace(0, 180, 180, endpoint=False) * degree

    # check x translation
    with pytest.raises(TypeError):
        nx_sample.x_translation = 56
    nx_sample.x_translation = numpy.linspace(0, 180, 180, endpoint=False) * meter

    # check y translation
    with pytest.raises(TypeError):
        nx_sample.y_translation = 56
    nx_sample.y_translation = [0.0] * 180 * meter

    # check z translation
    with pytest.raises(TypeError):
        nx_sample.z_translation = 56
    nx_sample.z_translation = None

    # check propagation distance
    with pytest.raises(TypeError):
        nx_sample.propagation_distance = "45"
    nx_sample.propagation_distance = None
    nx_sample.propagation_distance = 12.2 * ureg.centimeter

    # check pixel size
    with pytest.raises(TypeError):
        nx_sample.x_pixel_size = "toto"
    nx_sample.x_pixel_size = 12.6 * ureg.meter
    nx_sample.y_pixel_size = 5.6 * ureg.centimeter

    assert isinstance(nx_sample.to_nx_dict(), dict)

    # check we can't set undefined attributes
    with pytest.raises(AttributeError):
        nx_sample.test = 12

    # test concatenation
    nx_sample_concat = NXsample.concatenate([nx_sample, nx_sample])
    assert nx_sample_concat.name == "my sample"
    expected_rotation = (
        numpy.concatenate(
            [
                numpy.linspace(0, 180, 180, endpoint=False),
                numpy.linspace(0, 180, 180, endpoint=False),
            ]
        )
        * degree
    )
    numpy.testing.assert_array_equal(
        nx_sample_concat.rotation_angle.magnitude,
        expected_rotation.magnitude,
    )
    assert str(nx_sample_concat.rotation_angle.units) == str(expected_rotation.units)

    expected_x_translation = (
        numpy.concatenate(
            [
                numpy.linspace(0, 180, 180, endpoint=False),
                numpy.linspace(0, 180, 180, endpoint=False),
            ]
        )
        * meter
    )
    numpy.testing.assert_array_equal(
        nx_sample_concat.x_translation.magnitude,
        expected_x_translation.magnitude,
    )
    assert str(nx_sample_concat.x_translation.units) == str(
        expected_x_translation.units
    )

    expected_y_translation = (
        numpy.concatenate(
            [
                numpy.asarray([0.0] * 180),
                numpy.asarray([0.0] * 180),
            ]
        )
        * meter
    )
    numpy.testing.assert_array_equal(
        nx_sample_concat.y_translation.magnitude,
        expected_y_translation.magnitude,
    )
    assert str(nx_sample_concat.y_translation.units) == str(
        expected_y_translation.units
    )

    assert nx_sample_concat.z_translation is None
    assert nx_sample_concat.x_pixel_size == nx_sample.x_pixel_size
    assert nx_sample_concat.y_pixel_size == nx_sample.y_pixel_size



"""
Module for handling an `NXmonitor <https://manual.nexusformat.org/classes/base_classes/NXmonitor.html>`_.
"""

from functools import partial
from operator import is_not

import numpy
import pint
from silx.utils.proxy import docstring

from nxtomo.nxobject.nxobject import NXobject
from nxtomo.paths.nxtomo import get_paths as get_nexus_paths
from nxtomo.utils import get_quantity

__all__ = [
    "NXmonitor",
]

_ureg = pint.UnitRegistry()
_ampere = _ureg.ampere


class NXmonitor(NXobject):
    def __init__(self, node_name="control", parent: NXobject | None = None) -> None:
        """
        Representation of `NeXus NXmonitor <https://manual.nexusformat.org/classes/base_classes/NXmonitor.html>`_.
        A monitor of incident beam data.

        :param node_name: name of the monitor in the hierarchy.
        :param parent: parent in the NeXus hierarchy.
        """
        super().__init__(node_name=node_name, parent=parent)
        self._set_freeze(False)
        self._data = None
        self._set_freeze(True)

    @property
    def data(self) -> pint.Quantity | None:
        """
        Monitor data.
        In the case of NXtomo, it is expected to contain the machine electric current for each frame.
        """
        return self._data

    @data.setter
    def data(self, data: pint.Quantity | numpy.ndarray | list | tuple | None):
        if isinstance(data, pint.Quantity):
            self._data = data.to(_ampere)
        elif isinstance(data, (tuple, list)):
            if len(data) == 0:
                self._data = None
            else:
                self._data = numpy.asarray(data) * _ampere
        elif isinstance(data, numpy.ndarray):
            if data.ndim != 1:
                raise ValueError(f"data is expected to be 1D and not {data.ndim}D")
            self._data = data * _ampere
        elif data is None:
            self._data = None
        else:
            raise TypeError(
                f"data is expected to be a pint.Quantity, None, a list, or a 1D numpy array. Not {type(data)}"
            )

    @docstring(NXobject)
    def to_nx_dict(
        self,
        nexus_path_version: float | None = None,
        data_path: str | None = None,
    ) -> dict:
        nexus_paths = get_nexus_paths(nexus_path_version)
        monitor_nexus_paths = nexus_paths.nx_monitor_paths

        nx_dict = {}
        if self.data is not None:
            if monitor_nexus_paths.DATA_PATH is not None:
                data_path = f"{self.path}/{monitor_nexus_paths.DATA_PATH}"
                nx_dict[data_path] = self.data.magnitude
                nx_dict["@".join([data_path, "units"])] = f"{self.data.units:~}"

        if nx_dict != {}:
            nx_dict[f"{self.path}@NX_class"] = "NXmonitor"
        return nx_dict

    def _load(self, file_path: str, data_path: str, nexus_version: float) -> NXobject:
        """
        Create and load an NXmonitor from data on disk.
        """
        nexus_paths = get_nexus_paths(nexus_version)
        monitor_nexus_paths = nexus_paths.nx_monitor_paths
        if monitor_nexus_paths.DATA_PATH is not None:
            self.data = get_quantity(
                file_path=file_path,
                data_path="/".join([data_path, monitor_nexus_paths.DATA_PATH]),
                default_unit=_ampere,
            )

    @staticmethod
    @docstring(NXobject)
    def concatenate(nx_objects: tuple, node_name: str = "control"):
        # filter None obj
        nx_objects = tuple(filter(partial(is_not, None), nx_objects))
        if len(nx_objects) == 0:
            return None
        nx_monitor = NXmonitor(node_name=node_name)
        data = [nx_obj.data for nx_obj in nx_objects if nx_obj.data is not None]
        if len(data) > 0:
            nx_monitor.data = numpy.concatenate(data)
        return nx_monitor



"""Helper classes to define transformations contained in NXtransformations."""

import logging

import numpy
import pint
from silx.utils.enum import Enum as _Enum

# Create a pint unit registry
_ureg = pint.UnitRegistry()
_degree = _ureg.degree
_meter = _ureg.meter
_radian = _ureg.radian
_second = _ureg.second

_logger = logging.getLogger(__name__)

__all__ = [
    "TransformationType",
    "TransformationAxis",
    "Transformation",
    "DetXFlipTransformation",
    "DetYFlipTransformation",
    "DetZFlipTransformation",
    "GravityTransformation",
    "get_lr_flip",
    "get_ud_flip",
    "build_matrix",
]


class TransformationType(_Enum):
    """
    Possible NXtransformations types.
    """

    TRANSLATION = "translation"
    ROTATION = "rotation"
    GRAVITY = "gravity"


class TransformationAxis:
    """
    Predefined axes for tomography acquisitions performed at ESRF.
    Warning: these are stored as (X, Y, Z) and not under the usual NumPy reference (Z, Y, X).

    See: https://tomo.gitlab-pages.esrf.fr/ebs-tomo/master/modelization.html
    """

    AXIS_X = (1, 0, 0)
    AXIS_Y = (0, 1, 0)
    AXIS_Z = (0, 0, 1)


class Transformation:
    """
    Define a transformation applied along an axis.

    :param axis_name: name of the transformation.
    :param transformation_type: type of the transformation (the type should not change once created).
    :param vector: transformation vector as a tuple of three values or an instance of TransformationAxis.
    :param depends_on: name of another transformation on which this transformation depends.
    .. warning:: For rotations, values given in radians are automatically converted to degrees when comparing.
    """

    __isfrozen = False

    def __init__(
        self,
        axis_name: str,
        value: pint.Quantity | None,
        transformation_type: TransformationType,
        vector: tuple[float, float, float] | TransformationAxis,
        depends_on: str | None = None,
    ) -> None:
        self._axis_name = axis_name
        # Set the transformation type first so the setter can use it.
        self._transformation_type = TransformationType(transformation_type)
        self._transformation_values: pint.Quantity | None = None
        self.transformation_values = value
        if isinstance(vector, TransformationAxis):
            self._vector = vector.value()
        elif not isinstance(vector, (tuple, list, numpy.ndarray)) or len(vector) != 3:
            raise TypeError(
                f"vector should be a tuple of three elements. {vector} provided"
            )
        else:
            self._vector = tuple(vector)
        self._offset = (0, 0, 0)
        self._depends_on = None
        self.depends_on = depends_on
        self._equipment_component = None
        self._set_freeze()

    def _set_freeze(self, freeze=True):
        self.__isfrozen = freeze

    @property
    def axis_name(self) -> str:
        return self._axis_name

    @axis_name.setter
    def axis_name(self, axis_name: str):
        self._axis_name = axis_name

    @property
    def transformation_values(self) -> pint.Quantity | None:
        return self._transformation_values

    @transformation_values.setter
    def transformation_values(self, values: None | pint.Quantity):
        if isinstance(values, pint.Quantity):
            valid_dimensionalities = (
                _meter.dimensionality,
                _degree.dimensionality,
                (_meter / _second**2).dimensionality,
            )
            if values.dimensionality not in valid_dimensionalities:
                raise TypeError(
                    f"Unsupported dimensionality. Got {values.dimensionality} when should in {valid_dimensionalities}"
                )
        elif values is not None:
            raise TypeError(
                f"'values' is expected to be None or a pint.Quantity. Got {type(values)}"
            )
        self._transformation_values = values

    @property
    def transformation_type(self) -> TransformationType:
        return self._transformation_type

    @property
    def vector(self) -> tuple[float, float, float]:
        return self._vector

    @property
    def offset(self) -> tuple:
        return self._offset

    @offset.setter
    def offset(self, offset: tuple | list | numpy.ndarray):
        if not isinstance(offset, (tuple, list, numpy.ndarray)):
            raise TypeError(
                f"offset is expected to be a vector of three elements. {type(offset)} provided"
            )
        elif not len(offset) == 3:
            raise TypeError(
                f"offset is expected to be a vector of three elements. {offset} provided"
            )
        self._offset = tuple(offset)

    @property
    def depends_on(self):
        return self._depends_on

    @depends_on.setter
    def depends_on(self, depends_on):
        if not (depends_on is None or isinstance(depends_on, str)):
            raise TypeError(
                f"depends_on is expected to be None or str. {type(depends_on)} provided"
            )
        self._depends_on = depends_on

    @property
    def equipment_component(self) -> str | None:
        return self._equipment_component

    @equipment_component.setter
    def equipment_component(self, equipment_component: str | None):
        if not (equipment_component is None or isinstance(equipment_component, str)):
            raise TypeError(
                f"equipment_component is expected to be None or a str. {type(equipment_component)} provided"
            )
        self._equipment_component = equipment_component

    def to_nx_dict(self, transformations_nexus_paths, data_path: str):
        def join(my_list):
            my_list = tuple(filter(bool, my_list))
            return "" if len(my_list) == 0 else "/".join(my_list)

        quantity = self.transformation_values
        if quantity is None:
            _logger.error(f"no values defined for {self.axis_name}")
            transformation_values = None
            unit_str = ""
        else:
            if self.transformation_type is TransformationType.ROTATION:
                export_unit = _degree
            elif self.transformation_type is TransformationType.TRANSLATION:
                export_unit = _meter
            elif self.transformation_type is TransformationType.GRAVITY:
                export_unit = _meter / _second**2
            else:
                export_unit = quantity.units
                quantity_converted = quantity

            quantity_converted = quantity.to(export_unit)
            unit_str = f"{export_unit:~}"

            transformation_values = quantity_converted.magnitude

        res = {
            join((data_path, self.axis_name)): transformation_values,
            join(
                (
                    data_path,
                    self.axis_name + transformations_nexus_paths.TRANSFORMATION_TYPE,
                )
            ): self.transformation_type.value,
            join((data_path, f"{self.axis_name}@units")): unit_str,
        }

        # vector is mandatory
        res[
            join((data_path, f"{self.axis_name}{transformations_nexus_paths.VECTOR}"))
        ] = self.vector
        if self.offset is not None:
            res[
                join(
                    (data_path, f"{self.axis_name}{transformations_nexus_paths.OFFSET}")
                )
            ] = self.offset
        if self.depends_on:
            res[
                join(
                    (
                        data_path,
                        f"{self.axis_name}{transformations_nexus_paths.DEPENDS_ON}",
                    )
                )
            ] = self.depends_on
        if self.equipment_component:
            res[
                join(
                    (
                        data_path,
                        f"{self.axis_name}{transformations_nexus_paths.EQUIPMENT_COMPONENT}",
                    )
                )
            ] = self.equipment_component
        return res

    @staticmethod
    def from_nx_dict(axis_name: str, dict_: dict, transformations_nexus_paths):
        if transformations_nexus_paths is None:
            _logger.warning(
                "no transformations_nexus_paths (not implemented on this version of nexus - too old)"
            )
            return None
        value = dict_.get(axis_name, None)
        if isinstance(value, numpy.ndarray) and value.ndim == 0:
            value = value[()]
        vector = dict_.get(f"{axis_name}{transformations_nexus_paths.VECTOR}", None)
        transformation_type = dict_.get(
            f"{axis_name}{transformations_nexus_paths.TRANSFORMATION_TYPE}", None
        )
        if vector is None or transformation_type is None:
            raise ValueError(
                "Unable to find mandatory vector and/or transformation_type"
            )
        transformation_type = TransformationType(transformation_type)

        units_str = dict_.get(f"{axis_name}@units", None) or dict_.get(
            f"{axis_name}@unit", None
        )
        if units_str is not None and value is not None:
            if units_str == "m/s2":
                # backward with nxtomo < 2.0 (nxtomo writer version == 1.4)
                # note: the unit was typed differently, not recognized by pint
                # and the transformation type (gravity) was not existing
                units_str = "m / s ** 2"
                if transformation_type is TransformationType.TRANSLATION:
                    transformation_type = TransformationType.GRAVITY

            value = value * _ureg(units_str)

        transformation = Transformation(
            axis_name=axis_name,
            value=value,
            transformation_type=transformation_type,
            vector=vector,
        )

        offset = dict_.get(f"{axis_name}{transformations_nexus_paths.OFFSET}", None)
        if offset is not None:
            transformation.offset = offset

        depends_on = dict_.get(
            f"{axis_name}{transformations_nexus_paths.DEPENDS_ON}", None
        )
        if depends_on is not None:
            transformation.depends_on = depends_on

        equipment_component = dict_.get(
            f"{axis_name}{transformations_nexus_paths.EQUIPMENT_COMPONENT}", None
        )
        if equipment_component is not None:
            transformation.equipment_component = equipment_component

        return transformation

    def __setattr__(self, name, value):
        if self.__isfrozen and not hasattr(self, name):
            raise AttributeError("can't set attribute", name)
        else:
            super().__setattr__(name, value)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Transformation):
            return False
        same_dependence = self._depends_on == other.depends_on or (
            self._depends_on in (None, GravityTransformation(), "gravity")
            and other._depends_on in (None, GravityTransformation(), "gravity")
        )
        if not (
            self.vector == other.vector
            and self.transformation_type == other.transformation_type
            and self.offset == other.offset
            and same_dependence
            and self.equipment_component == other.equipment_component
        ):
            return False

        q1 = self.transformation_values
        q2 = other.transformation_values
        if q1 is None or q2 is None:
            return q1 is q2
        try:
            if self.transformation_type is TransformationType.GRAVITY:
                v1 = q1.to(_ureg("m/s^2")).magnitude
                v2 = q2.to(_ureg("m/s^2")).magnitude
            if self.transformation_type is TransformationType.ROTATION:
                v1 = q1.to(_degree).magnitude % 360
                v2 = q2.to(_degree).magnitude % 360
            elif self.transformation_type is TransformationType.TRANSLATION:
                v1 = q1.to(_meter).magnitude
                v2 = q2.to(_meter).magnitude
            else:
                v1 = q1.magnitude
                v2 = q2.to(q1.units).magnitude
        except Exception:
            return False

        if isinstance(v1, numpy.ndarray) or isinstance(v2, numpy.ndarray):
            return numpy.array_equal(v1, v2)
        else:
            return v1 == v2

    def as_matrix(self):
        if self.transformation_values is None:
            raise ValueError(f"missing transformation values for {self}")
        # Use the magnitude from the pint.Quantity
        if numpy.isscalar(self.transformation_values.magnitude):
            if self.transformation_type is TransformationType.ROTATION:
                theta = self.transformation_values.to(_radian).magnitude
                if self.offset != (0, 0, 0):
                    raise ValueError("offset not handled")
                if self.vector == (1, 0, 0):
                    return numpy.array(
                        [
                            [1, 0, 0],
                            [0, numpy.cos(theta), -numpy.sin(theta)],
                            [0, numpy.sin(theta), -numpy.cos(theta)],
                        ],
                        dtype=numpy.float32,
                    )
                elif self.vector == (0, 1, 0):
                    return numpy.array(
                        [
                            [numpy.cos(theta), 0, numpy.sin(theta)],
                            [0, 1, 0],
                            [-numpy.sin(theta), 0, numpy.cos(theta)],
                        ],
                        dtype=numpy.float32,
                    )
                elif self.vector == (0, 0, 1):
                    return numpy.array(
                        [
                            [numpy.cos(theta), -numpy.sin(theta), 0],
                            [numpy.sin(theta), numpy.cos(theta), 0],
                            [0, 0, 1],
                        ],
                        dtype=numpy.float32,
                    )
                else:
                    raise ValueError(f"vector {self.vector} not handled")
            elif self.transformation_type is TransformationType.TRANSLATION:
                val = self.transformation_values.to(_meter).magnitude
                if self.vector == (1, 0, 0):
                    return numpy.array(
                        [
                            [val, 0, 0],
                            [0, 1, 0],
                            [0, 0, 1],
                        ],
                        dtype=numpy.float32,
                    )
                elif self.vector == (0, 1, 0):
                    return numpy.array(
                        [
                            [1, 0, 0],
                            [0, val, 0],
                            [0, 0, 1],
                        ],
                        dtype=numpy.float32,
                    )
                elif self.vector == (0, 0, 1):
                    return numpy.array(
                        [
                            [1, 0, 0],
                            [0, 1, 0],
                            [0, 0, val],
                        ],
                        dtype=numpy.float32,
                    )
            else:
                raise RuntimeError(
                    f"unknown transformation type: {self.transformation_type}"
                )
        else:
            raise ValueError(
                f"transformations as a list of values is not handled for now ({self})"
            )

    def __str__(self):
        return f"transformation: {self.axis_name} -" + ", ".join(
            [
                f"type: {self.transformation_type.value}",
                f"value: {self.transformation_values}",
                f"vector: {self.vector}",
                f"offset: {self.offset}",
                f"depends_on: {self.depends_on}",
                f"equipment_component: {self.equipment_component}",
            ]
        )


class DetXFlipTransformation(Transformation):
    """
    Convenient class to a flip with X (1, 0, 0) as the rotation axis.
    """

    def __init__(self, flip: bool, axis_name="rx", depends_on=None) -> None:
        value = 180 if flip else 0
        super().__init__(
            axis_name=axis_name,
            value=value * _degree,
            transformation_type=TransformationType.ROTATION,
            vector=TransformationAxis.AXIS_X,
            depends_on=depends_on,
        )


class DetYFlipTransformation(Transformation):
    """
    Convenient class to a flip with Y (0, 1, 0) as the rotation axis.
    """

    def __init__(self, flip: bool, axis_name="ry", depends_on=None) -> None:
        value = 180 if flip else 0
        super().__init__(
            axis_name=axis_name,
            value=value * _degree,
            transformation_type=TransformationType.ROTATION,
            vector=TransformationAxis.AXIS_Y,
            depends_on=depends_on,
        )


class DetZFlipTransformation(Transformation):
    """
    Convenient class to a flip with Z (0, 0, 1) as the rotation axis.
    """

    def __init__(self, flip: bool, axis_name="rz", depends_on=None) -> None:
        value = 180 if flip else 0
        super().__init__(
            axis_name=axis_name,
            value=value * _degree,
            transformation_type=TransformationType.ROTATION,
            vector=TransformationAxis.AXIS_Z,
            depends_on=depends_on,
        )


class GravityTransformation(Transformation):
    """
    Gravity is used to solve the transformation chain (acting as the chain endpoint).
    The direction is set to -Z, and the dimension is unitless because it is used to
    resolve the transformation chain.
    """

    def __init__(self) -> None:
        super().__init__(
            axis_name="gravity",
            value=9.80665 * (_meter / _second**2),
            transformation_type=TransformationType.GRAVITY,
            vector=(0, 0, -1),
        )


def get_lr_flip(transformations: tuple) -> tuple:
    """
    Check all transformations for those matching a left-right detector flip and return them.
    """
    if not isinstance(transformations, (tuple, list)):
        raise TypeError(
            f"transformations is expected to be a tuple. {type(transformations)} provided"
        )
    res = []
    for transformation in transformations:
        if transformation in (
            DetYFlipTransformation(flip=True),
            DetYFlipTransformation(flip=False),
        ):
            res.append(transformation)
    return tuple(res)


def get_ud_flip(transformations: tuple) -> tuple:
    """
    Check all transformations for those matching an up-down detector flip and return them.
    """
    if not isinstance(transformations, (tuple, list)):
        raise TypeError(
            f"transformations is expected to be a tuple. {type(transformations)} provided"
        )
    res = []
    for transformation in transformations:
        if transformation in (
            DetXFlipTransformation(flip=True),
            DetXFlipTransformation(flip=False),
        ):
            res.append(transformation)
    return tuple(res)


def build_matrix(transformations: set):
    """
    Build a matrix from a set of Transformations.
    """
    transformations = {
        transformation.axis_name: transformation for transformation in transformations
    }
    already_applied_transformations = set(["gravity"])

    def handle_transformation(transformation: Transformation, matrix):
        if not isinstance(transformation, Transformation):
            raise TypeError(
                f"transformation is expected to be an instance of Transformation. {type(transformation)} provided"
            )
        # Handle dependencies
        if transformation.axis_name in already_applied_transformations:
            return matrix
        elif transformation.transformation_values is None:
            if transformation.axis_name.lower() == "gravity":
                return numpy.identity(3, dtype=numpy.float32)
            else:
                _logger.error(
                    f"transformation value not provided for {transformation.axis_name}. Ignoring transformation"
                )
                return matrix
        elif (
            transformation.depends_on is not None
            and transformation.depends_on not in already_applied_transformations
        ):
            if transformation.depends_on not in transformations:
                raise ValueError(
                    f"Unable to find transformation {transformation.depends_on}. "
                    "Broken dependency chain."
                )
            else:
                matrix = handle_transformation(
                    transformations[transformation.depends_on], matrix
                )
        matrix = numpy.matmul(matrix, transformation.as_matrix())
        already_applied_transformations.add(transformation.axis_name)
        return matrix

    matrix = numpy.identity(3, dtype=numpy.float32)
    for transformation in transformations.values():
        matrix = handle_transformation(transformation, matrix)
    return matrix



"""General utility helpers."""

import os

import h5py
import numpy
import pint
from silx.io.utils import h5py_read_dataset
from silx.io.utils import open as hdf5_open

from nxtomo.io import to_target_rel_path

_ureg = pint.UnitRegistry()

try:
    import tifffile  # noqa F401
except ImportError:
    has_tifffile = False
else:
    from tifffile import TiffFile

    has_tifffile = True

__all__ = [
    "cast_and_check_array_1D",
    "get_data",
    "get_quantity",
]


def cast_and_check_array_1D(array, array_name: str):
    """
    Cast the provided array to 1D and handle pint quantities.

    :param array: array to be cast to 1D. Can be a pint.Quantity.
    :param array_name: name of the array (used for logging only).
    :return: 1D NumPy array (with magnitude if a pint.Quantity).
    :raises TypeError: if the input type is invalid.
    :raises ValueError: if the input has more than one dimension.
    """
    if not isinstance(array, (type(None), numpy.ndarray, tuple, list, pint.Quantity)):
        raise TypeError(
            f"{array_name} is expected to be None, a numpy.ndarray, a tuple, a list, or a pint.Quantity. "
            f"Not {type(array)}"
        )
    if isinstance(array, pint.Quantity):
        unit = array.units
        array = array.magnitude
    else:
        unit = None

    if array is not None and not isinstance(array, numpy.ndarray):
        array = numpy.asarray(array)

    if array is not None and array.ndim > 1:
        raise ValueError(f"{array_name} is expected to be 0 or 1D, not {array.ndim}D")
    return array * unit if unit else array


def get_quantity(
    file_path: str, data_path: str, default_unit: pint.Unit
) -> pint.Quantity:
    """
    Return the value and unit of an HDF5 dataset. If the unit is not found, fall back on `default_unit`.

    :param file_path: file path location of the HDF5 dataset to read.
    :param data_path: data path location of the HDF5 dataset to read.
    :param default_unit: default unit to use if the dataset has no ``unit`` or ``units`` attribute, as a pint.Unit object.
    :return: pint.Quantity with the data and associated unit.
    """
    with hdf5_open(file_path) as h5f:
        if data_path in h5f and isinstance(h5f[data_path], h5py.Dataset):
            dataset = h5f[data_path]
            unit = None
            if "unit" in dataset.attrs:
                unit = dataset.attrs["unit"]
            elif "units" in dataset.attrs:
                unit = dataset.attrs["units"]
            else:
                unit = str(default_unit)  # Use default unit if none found
            if hasattr(unit, "decode"):
                unit = unit.decode()

            if unit == "kev":
                unit = "keV"
            try:
                unit = _ureg(unit)  # Convert to a pint unit
            except pint.UndefinedUnitError:
                raise ValueError(f"Invalid or undefined unit: {unit}")
            data = h5py_read_dataset(dataset)  # Read dataset values
            return data * unit
        else:
            return None


def get_data(file_path: str, data_path: str):
    """
    Proxy to `h5py_read_dataset`, handling the case where `data_path` is not present in the file.
    In this situation, return None.

    :param file_path: file path location of the HDF5 dataset to read.
    :param data_path: data path location of the HDF5 dataset to read.
    """
    with hdf5_open(file_path) as h5f:
        if data_path in h5f:
            return h5py_read_dataset(h5f[data_path])
        else:
            return None


def create_detector_dataset_from_tiff(
    tiff_files: tuple,
    external_dataset_group: h5py.Group,
    external_dataset_prefix="frame_",
    dtype=None,
    relative_link: bool = True,
) -> tuple[h5py.VirtualSource]:
    """
    Create a series of external datasets to TIFF files (one per file) inside ``external_dataset_group``.

    :param tiff_files: files used to create virtual sources.
    :param external_dataset_group: output HDF5 group. The file must be opened with write access (mode 'w' or 'a').
    :param dtype: expected dtype of all the TIFF data. If not provided it is deduced from the first dataset.
    :param relative_link: if True create the link using a relative path; otherwise use an absolute path.

    .. warning::

        The most robust way to create an NXtomo is to use relative links (in order to share it with the ``.tif`` files).
        Nevertheless there is currently a limitation on how relative links are resolved for external datasets
        (resolution is done according to the current working directory instead of the file).
        Tomotools handles this, but other software might not (such as silx, as this is a workaround that should be handled at the HDF5 level).
        Links might therefore appear broken when using relative paths, which will not happen when using absolute paths.
    """
    if not has_tifffile:
        raise RuntimeError("tiff file not installed")
    external_datasets = []

    # convert from local to ...

    for i_file, tiff_file in enumerate(tiff_files):
        with TiffFile(tiff_file, mode="r") as tif:
            fh = tif.filehandle
            for page in tif.pages:
                if dtype is not None:
                    assert dtype == page.dtype, "incoherent data type"
                dtype = page.dtype
                for index, (offset, bytecount) in enumerate(
                    zip(page.dataoffsets, page.databytecounts)
                ):
                    _ = fh.seek(offset)
                    data = fh.read(bytecount)
                    _, _, shape = page.decode(data, index, jpegtables=page.jpegtables)
                    if len(shape) == 4:
                        # don't know why but return it as 4D when 2D expected...
                        shape = shape[0:-1]
                    elif len(shape) == 2:
                        shape = 1, *shape

                    # move tiff file path to relative path
                    if relative_link:
                        external_file_path = to_target_rel_path(
                            file_path=tiff_file,
                            target_path=external_dataset_group.file.filename,
                        )
                    else:
                        external_file_path = os.path.abspath(tiff_file)

                    external_dataset = external_dataset_group.create_dataset(
                        name=f"{external_dataset_prefix}{str(i_file).zfill(6)}",
                        shape=shape,
                        dtype=dtype,
                        external=[(external_file_path, offset, bytecount)],
                    )
                    external_datasets.append(external_dataset)

    virtual_sources = []
    for i, ed in enumerate(external_datasets):
        vsource = h5py.VirtualSource(ed)
        virtual_sources.append(vsource)
    return tuple(virtual_sources)



"""NeXus paths used to define an `NXmonitor <https://manual.nexusformat.org/classes/base_classes/NXmonitor.html>`_."""


class NEXUS_MONITOR_PATH:
    DATA_PATH = "data"


class NEXUS_MONITOR_PATH_V_1_0(NEXUS_MONITOR_PATH):
    pass


class NEXUS_MONITOR_PATH_V_1_1(NEXUS_MONITOR_PATH_V_1_0):
    pass


class NEXUS_MONITOR_PATH_V_1_2(NEXUS_MONITOR_PATH_V_1_1):
    pass


class NEXUS_MONITOR_PATH_V_1_3(NEXUS_MONITOR_PATH_V_1_2):
    pass


class NEXUS_MONITOR_PATH_V_1_4(NEXUS_MONITOR_PATH_V_1_3):
    pass


class NEXUS_MONITOR_PATH_V_1_5(NEXUS_MONITOR_PATH_V_1_4):
    pass



import os
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
from datetime import datetime

import h5py
import numpy
import pint
import pytest
from silx.io.url import DataUrl
from silx.io.utils import h5py_read_dataset

from nxtomo.application.nxtomo import NXtomo, copy_nxtomo_file
from nxtomo.geometry._CoordinateSystem import CoordinateSystem
from nxtomo.io import HDF5File
from nxtomo.nxobject.nxdetector import FieldOfView, ImageKey
from nxtomo.nxobject.utils.concatenate import concatenate

try:
    import tifffile
except ImportError:
    has_tifffile = False
else:
    from nxtomo.utils.utils import create_detector_dataset_from_tiff

    has_tifffile = True


nexus_path_versions = (1.5, 1.4, 1.3, 1.2, 1.1, 1.0, None)
ureg = pint.get_application_registry()
second = ureg.second
degree = ureg.degree
meter = ureg.meter


@pytest.mark.parametrize("nexus_path_version", nexus_path_versions)
def test_nx_tomo(nexus_path_version, tmp_path):
    nx_tomo = NXtomo()

    # check start time
    with pytest.raises(TypeError):
        nx_tomo.start_time = 12
    nx_tomo.start_time = datetime.now()

    # check end time
    with pytest.raises(TypeError):
        nx_tomo.end_time = 12
    nx_tomo.end_time = datetime(2022, 2, 27)

    # check sample
    with pytest.raises(TypeError):
        nx_tomo.sample = "tata"

    # check detector
    with pytest.raises(TypeError):
        nx_tomo.instrument.detector = "tata"

    # check energy
    with pytest.raises(TypeError):
        nx_tomo.energy = "tata"
    nx_tomo.energy = 12.3 * ureg.keV

    # check group size
    with pytest.raises(TypeError):
        nx_tomo.group_size = "tata"
    nx_tomo.group_size = 3

    # check title
    with pytest.raises(TypeError):
        nx_tomo.title = 12
    nx_tomo.title = "title"

    # check instrument
    with pytest.raises(TypeError):
        nx_tomo.instrument = "test"

    # check we can't set undefined attributes
    with pytest.raises(AttributeError):
        nx_tomo.test = 12

    # create detector for test
    projections = numpy.random.random(100 * 100 * 8).reshape([8, 100, 100])
    flats_1 = numpy.random.random(100 * 100 * 2).reshape([2, 100, 100])
    darks = numpy.random.random(100 * 100 * 3).reshape([3, 100, 100])
    flats_2 = numpy.random.random(100 * 100 * 2).reshape([2, 100, 100])
    alignment = numpy.random.random(100 * 100 * 1).reshape([1, 100, 100])

    n_frames = 3 + 2 + 8 + 2 + 1

    nx_tomo.instrument.detector.data = numpy.concatenate(
        [
            darks,
            flats_1,
            projections,
            flats_2,
            alignment,
        ]
    )

    nx_tomo.instrument.detector.image_key_control = numpy.concatenate(
        [
            [ImageKey.DARK_FIELD] * 3,
            [ImageKey.FLAT_FIELD] * 2,
            [ImageKey.PROJECTION] * 8,
            [ImageKey.FLAT_FIELD] * 2,
            [ImageKey.ALIGNMENT] * 1,
        ]
    )
    nx_tomo.instrument.detector.sequence_number = numpy.linspace(
        0, n_frames, n_frames, dtype=numpy.uint32
    )

    nx_tomo.instrument.detector.x_pixel_size = (
        nx_tomo.instrument.detector.y_pixel_size
    ) = (1e-7 * meter)
    nx_tomo.instrument.detector.distance = 0.2 * meter
    nx_tomo.instrument.detector.field_of_view = FieldOfView.HALF
    nx_tomo.instrument.detector.count_time = (
        numpy.concatenate(
            [
                [0.2] * 3,  # darks
                [0.1] * 2,  # flats 1
                [0.1] * 8,  # projections
                [0.1] * 2,  # flats 2
                [0.1] * 1,  # alignment
            ]
        )
        * second
    )

    # create sample for test
    nx_tomo.sample.name = "my sample"
    nx_tomo.sample.rotation_angle = (
        numpy.concatenate(
            [
                [0.0] * 3,  # darks
                [0.0] * 2,  # flats 1
                numpy.linspace(0, 180, num=8, endpoint=False),  # projections
                [180.0] * 2,  # flats 2
                [0.0],  # alignment
            ]
        )
        * degree
    )

    if nexus_path_version is None or nexus_path_version >= 1.4:
        # create source and detector for test
        nx_tomo.instrument.source.distance = 3.6 * meter
        nx_tomo.instrument.detector.y_rotation_axis_pixel_position = 1.1
    nx_tomo.instrument.detector.x_rotation_axis_pixel_position = 1.2

    nx_tomo.sample.x_translation = [0.6] * n_frames * meter
    nx_tomo.sample.y_translation = [0.2] * n_frames * meter
    nx_tomo.sample.z_translation = [0.1] * n_frames * meter
    nx_tomo.sample.x_pixel_size = 90.8 * ureg.nanometer
    nx_tomo.sample.y_pixel_size = 8.2 * ureg.micrometer

    assert nx_tomo.is_root is True
    assert nx_tomo.instrument.is_root is False
    assert (
        nx_tomo.root_path
        == nx_tomo.instrument.root_path
        == nx_tomo.instrument.detector.root_path
    )

    NXtomo.check_consistency(nx_tomo=nx_tomo, raises_error=True)

    folder = tmp_path / "test_folder"
    folder.mkdir()

    file_path = os.path.join(folder, "nexus_file.hdf5")

    nx_tomo.save(
        file_path=file_path,
        data_path="entry",
        nexus_path_version=nexus_path_version,
    )
    assert os.path.exists(file_path)

    # insure we can read it back
    scan = NXtomo().load(file_path, data_path="entry")
    assert (
        len(
            tuple(
                filter(
                    lambda image_key: image_key is ImageKey.FLAT_FIELD,
                    scan.instrument.detector.image_key_control,
                )
            )
        )
        == 4
    )
    assert (
        len(
            tuple(
                filter(
                    lambda image_key: image_key is ImageKey.DARK_FIELD,
                    scan.instrument.detector.image_key_control,
                )
            )
        )
        == 3
    )
    assert (
        len(
            tuple(
                filter(
                    lambda image_key: image_key is ImageKey.PROJECTION,
                    scan.instrument.detector.image_key_control,
                )
            )
        )
        == 8
    )
    assert (
        len(
            tuple(
                filter(
                    lambda image_key: image_key is ImageKey.ALIGNMENT,
                    scan.instrument.detector.image_key_control,
                )
            )
        )
        == 1
    )
    if nexus_path_version is None or nexus_path_version >= 1.5:
        numpy.testing.assert_array_equal(
            scan.instrument.detector.sequence_number,
            numpy.linspace(0, n_frames, n_frames, dtype=numpy.uint32),
        )
    assert scan.energy.magnitude == 12.3
    assert scan.instrument.detector.x_pixel_size.magnitude == 1e-7
    assert scan.instrument.detector.y_pixel_size.magnitude == 1e-7
    assert scan.instrument.detector.distance.magnitude == 0.2
    assert scan.instrument.detector.field_of_view == FieldOfView.HALF
    assert scan.sample.name == "my sample"
    assert (
        len(scan.sample.x_translation.magnitude)
        == len(scan.sample.y_translation.magnitude)
        == len(scan.sample.z_translation.magnitude)
        == n_frames
    )
    assert scan.sample.x_translation.magnitude[0] == 0.6
    assert scan.sample.y_translation.magnitude[0] == 0.2
    assert scan.sample.z_translation.magnitude[0] == 0.1
    if nexus_path_version != 1.0:
        assert scan.instrument.source.name is not None
        assert scan.instrument.source.type is not None
    if nexus_path_version is None or nexus_path_version >= 1.4:
        assert nx_tomo.instrument.source.distance.to_base_units().magnitude == 3.6
        assert nx_tomo.instrument.detector.y_rotation_axis_pixel_position == 1.1
    assert nx_tomo.instrument.detector.x_rotation_axis_pixel_position == 1.2

    # try to load it from the disk
    loaded_nx_tomo = NXtomo().load(file_path=file_path, data_path="entry")
    assert isinstance(loaded_nx_tomo, NXtomo)
    assert loaded_nx_tomo.energy.magnitude == nx_tomo.energy.magnitude
    assert str(loaded_nx_tomo.energy.units) == str(nx_tomo.energy.units)
    assert loaded_nx_tomo.start_time == nx_tomo.start_time
    assert loaded_nx_tomo.end_time == nx_tomo.end_time
    if nexus_path_version is None or nexus_path_version >= 1.4:
        assert (
            loaded_nx_tomo.instrument.source.distance.to_base_units().magnitude
            == nx_tomo.instrument.source.distance.to_base_units().magnitude
        )
        assert (
            loaded_nx_tomo.instrument.detector.y_rotation_axis_pixel_position
            == nx_tomo.instrument.detector.y_rotation_axis_pixel_position
        )
    assert (
        loaded_nx_tomo.instrument.detector.x_rotation_axis_pixel_position
        == nx_tomo.instrument.detector.x_rotation_axis_pixel_position
    )
    numpy.testing.assert_equal(
        loaded_nx_tomo.instrument.detector.x_pixel_size.to_base_units().magnitude,
        nx_tomo.instrument.detector.x_pixel_size.to_base_units().magnitude,
    )
    assert str(loaded_nx_tomo.instrument.detector.x_pixel_size.units) == str(
        nx_tomo.instrument.detector.x_pixel_size.units
    )
    numpy.testing.assert_equal(
        loaded_nx_tomo.instrument.detector.y_pixel_size.to_base_units().magnitude,
        nx_tomo.instrument.detector.y_pixel_size.to_base_units().magnitude,
    )
    assert str(loaded_nx_tomo.instrument.detector.y_pixel_size.units) == str(
        nx_tomo.instrument.detector.y_pixel_size.units
    )
    assert (
        loaded_nx_tomo.instrument.detector.field_of_view
        == nx_tomo.instrument.detector.field_of_view
    )
    numpy.testing.assert_equal(
        loaded_nx_tomo.instrument.detector.count_time.to_base_units().magnitude,
        nx_tomo.instrument.detector.count_time.to_base_units().magnitude,
    )
    assert str(loaded_nx_tomo.instrument.detector.count_time.units) == str(
        nx_tomo.instrument.detector.count_time.units
    )
    assert (
        loaded_nx_tomo.instrument.detector.distance
        == nx_tomo.instrument.detector.distance
    )

    numpy.testing.assert_array_equal(
        loaded_nx_tomo.instrument.detector.image_key_control,
        nx_tomo.instrument.detector.image_key_control,
    )
    numpy.testing.assert_array_equal(
        loaded_nx_tomo.instrument.detector.image_key,
        nx_tomo.instrument.detector.image_key,
    )
    if nexus_path_version is None or nexus_path_version >= 1.5:
        numpy.testing.assert_array_equal(
            loaded_nx_tomo.instrument.detector.sequence_number,
            nx_tomo.instrument.detector.sequence_number,
        )
    else:
        assert loaded_nx_tomo.instrument.detector.sequence_number is None
    assert loaded_nx_tomo.sample.name == nx_tomo.sample.name
    assert loaded_nx_tomo.sample.rotation_angle is not None
    numpy.testing.assert_array_almost_equal(
        loaded_nx_tomo.sample.rotation_angle.magnitude,
        nx_tomo.sample.rotation_angle.magnitude,
    )
    assert str(loaded_nx_tomo.sample.rotation_angle.units) == str(
        nx_tomo.sample.rotation_angle.units
    )
    numpy.testing.assert_array_almost_equal(
        loaded_nx_tomo.sample.x_translation.magnitude,
        nx_tomo.sample.x_translation.magnitude,
    )
    assert str(loaded_nx_tomo.sample.x_translation.units) == str(
        nx_tomo.sample.x_translation.units
    )
    numpy.testing.assert_array_almost_equal(
        loaded_nx_tomo.sample.y_translation.magnitude,
        nx_tomo.sample.y_translation.magnitude,
    )
    assert str(loaded_nx_tomo.sample.y_translation.units) == str(
        nx_tomo.sample.y_translation.units
    )
    numpy.testing.assert_array_almost_equal(
        loaded_nx_tomo.sample.z_translation.magnitude,
        nx_tomo.sample.z_translation.magnitude,
    )
    assert str(loaded_nx_tomo.sample.z_translation.units) == str(
        nx_tomo.sample.z_translation.units
    )
    if nexus_path_version is None or nexus_path_version >= 1.5:
        assert loaded_nx_tomo.sample.x_pixel_size == 90.8 * ureg.nanometer
        assert loaded_nx_tomo.sample.y_pixel_size == 8.2 * ureg.micrometer
    else:
        assert loaded_nx_tomo.sample.x_pixel_size is None
        assert loaded_nx_tomo.sample.y_pixel_size is None

    loaded_nx_tomo = NXtomo().load(
        file_path=file_path, data_path="entry", detector_data_as="as_numpy_array"
    )
    numpy.testing.assert_array_almost_equal(
        loaded_nx_tomo.instrument.detector.data,
        nx_tomo.instrument.detector.data,
    )
    loaded_nx_tomo = NXtomo().load(
        file_path=file_path, data_path="entry", detector_data_as="as_data_url"
    )
    assert isinstance(loaded_nx_tomo.instrument.detector.data[0], DataUrl)
    with pytest.raises(ValueError):
        # check an error is raise because the dataset is not virtual
        loaded_nx_tomo = NXtomo().load(
            file_path=file_path,
            data_path="entry",
            detector_data_as="as_virtual_source",
        )
    assert loaded_nx_tomo._coordinate_system == CoordinateSystem.McStas

    # test concatenation
    nx_tomo_concat = concatenate([loaded_nx_tomo, None, loaded_nx_tomo])
    concat_file = os.path.join(folder, "concatenated_nexus_file.hdf5")

    nx_tomo_concat.save(
        file_path=concat_file,
        data_path="myentry",
        nexus_path_version=nexus_path_version,
    )
    loaded_concatenated_nx_tomo = NXtomo().load(
        file_path=concat_file,
        data_path="myentry",
        detector_data_as="as_virtual_source",
    )
    expected_rotation_angles = numpy.concatenate(
        [
            nx_tomo.sample.rotation_angle,
            nx_tomo.sample.rotation_angle,
        ]
    )
    numpy.testing.assert_array_almost_equal(
        loaded_concatenated_nx_tomo.sample.rotation_angle.magnitude,
        expected_rotation_angles.magnitude,
    )
    assert str(loaded_concatenated_nx_tomo.sample.rotation_angle.units) == str(
        expected_rotation_angles.units
    )
    expected_x_translation = numpy.concatenate(
        [
            nx_tomo.sample.x_translation,
            nx_tomo.sample.x_translation,
        ]
    )
    numpy.testing.assert_array_almost_equal(
        loaded_concatenated_nx_tomo.sample.x_translation.magnitude,
        expected_x_translation.magnitude,
    )
    assert str(loaded_concatenated_nx_tomo.sample.x_translation.units) == str(
        expected_x_translation.units
    )

    with pytest.raises(TypeError):
        concatenate([1, 2])

    with h5py.File(concat_file, mode="r") as h5f:
        h5py_read_dataset(h5f["myentry/definition"]) == "NXtomo"

    if nexus_path_version is None or nexus_path_version >= 1.4:
        assert (
            loaded_concatenated_nx_tomo.instrument.source.distance.to_base_units().magnitude
            == loaded_nx_tomo.instrument.source.distance.to_base_units().magnitude
        )
        assert (
            loaded_concatenated_nx_tomo.instrument.detector.x_rotation_axis_pixel_position
            == loaded_nx_tomo.instrument.detector.x_rotation_axis_pixel_position
        )
        assert (
            loaded_concatenated_nx_tomo.instrument.detector.y_rotation_axis_pixel_position
            == loaded_nx_tomo.instrument.detector.y_rotation_axis_pixel_position
        )


@pytest.mark.parametrize("nexus_path_version", nexus_path_versions)
def test_nx_tomo_subselection(nexus_path_version):
    """
    test sub_select_from_projection_angle_range
    """
    nx_tomo = NXtomo()
    nx_tomo.energy = 12.3 * ureg.keV
    shape = (12, 12)
    data_dark = numpy.ones(shape)
    data_flat = numpy.ones(shape) * 2.0
    data_projection = numpy.ones(shape) * 3.0
    str(nx_tomo)
    nx_tomo.instrument.detector.data = numpy.concatenate(
        (
            data_dark,
            data_dark,
            data_flat,
            data_projection,
            data_projection,
            data_projection,
            data_flat,
            data_projection,
            data_projection,
            data_projection,
            data_flat,
        )
    )
    nx_tomo.instrument.detector.image_key_control = numpy.array(
        (
            ImageKey.DARK_FIELD,
            ImageKey.DARK_FIELD,
            ImageKey.FLAT_FIELD,
            ImageKey.PROJECTION,
            ImageKey.PROJECTION,
            ImageKey.PROJECTION,
            ImageKey.FLAT_FIELD,
            ImageKey.PROJECTION,
            ImageKey.PROJECTION,
            ImageKey.PROJECTION,
            ImageKey.FLAT_FIELD,
        )
    )
    nx_tomo.instrument.detector.sequence_number = numpy.linspace(
        0, 11, 11, dtype=numpy.uint32
    )

    original_angles = numpy.array(
        (
            0,
            0,
            0,
            10,
            20.5,
            22.5,
            180,
            180,
            200,
            300.2,
            300.2,
        )
    )
    nx_tomo.sample.rotation_angle = original_angles * degree

    nx_tomo_sub_1 = NXtomo.sub_select_from_angle_offset(
        nx_tomo=nx_tomo,
        start_angle_offset=10,
        angle_interval=12.5,
        shift_angles=False,
    )

    numpy.testing.assert_equal(
        nx_tomo_sub_1.instrument.detector.image_key_control,
        numpy.array(
            (
                ImageKey.DARK_FIELD,
                ImageKey.DARK_FIELD,
                ImageKey.FLAT_FIELD,
                ImageKey.INVALID,
                ImageKey.PROJECTION,
                ImageKey.PROJECTION,
                ImageKey.FLAT_FIELD,
                ImageKey.INVALID,
                ImageKey.INVALID,
                ImageKey.INVALID,
                ImageKey.FLAT_FIELD,
            )
        ),
    )
    numpy.testing.assert_equal(
        nx_tomo_sub_1.sample.rotation_angle.magnitude,
        original_angles,
    )
    assert str(nx_tomo_sub_1.sample.rotation_angle.units) == str(degree)

    nx_tomo_sub_2 = NXtomo.sub_select_from_angle_offset(
        nx_tomo=nx_tomo,
        start_angle_offset=10,
        angle_interval=20,
        shift_angles=True,
    )

    numpy.testing.assert_equal(
        nx_tomo_sub_2.sample.rotation_angle.magnitude[0:3],
        0.0,
    )
    numpy.testing.assert_array_equal(
        nx_tomo_sub_2.sample.rotation_angle.magnitude[3:6],
        numpy.array([0.0, 10.5, 12.5]),
    )
    assert str(nx_tomo_sub_2.sample.rotation_angle.units) == str(degree)

    nx_tomo_sub_3 = NXtomo.sub_select_from_angle_offset(
        nx_tomo=nx_tomo,
        start_angle_offset=-10,
        angle_interval=300,
        shift_angles=False,
    )

    numpy.testing.assert_equal(
        nx_tomo_sub_3.instrument.detector.image_key_control,
        numpy.array(
            (
                ImageKey.DARK_FIELD,
                ImageKey.DARK_FIELD,
                ImageKey.FLAT_FIELD,
                ImageKey.PROJECTION,
                ImageKey.PROJECTION,
                ImageKey.PROJECTION,
                ImageKey.FLAT_FIELD,
                ImageKey.PROJECTION,
                ImageKey.PROJECTION,
                ImageKey.INVALID,
                ImageKey.FLAT_FIELD,
            )
        ),
    )

    nx_tomo_sub_4 = NXtomo.sub_select_from_angle_offset(
        nx_tomo=nx_tomo,
        start_angle_offset=-10,
        angle_interval=None,
        shift_angles=False,
    )
    numpy.testing.assert_equal(
        nx_tomo_sub_4.instrument.detector.image_key_control,
        numpy.array(
            (
                ImageKey.DARK_FIELD,
                ImageKey.DARK_FIELD,
                ImageKey.FLAT_FIELD,
                ImageKey.PROJECTION,
                ImageKey.PROJECTION,
                ImageKey.PROJECTION,
                ImageKey.FLAT_FIELD,
                ImageKey.PROJECTION,
                ImageKey.PROJECTION,
                ImageKey.PROJECTION,
                ImageKey.FLAT_FIELD,
            )
        ),
    )
    numpy.testing.assert_equal(
        nx_tomo_sub_4.instrument.detector.sequence_number,
        numpy.linspace(0, 11, 11, dtype=numpy.uint32),
    )


def test_bliss_original_files(tmp_path):
    """
    test about NXtomo.bliss_original_files
    """
    test_dir = tmp_path / "test_bliss_original_files"
    test_dir.mkdir()

    nx_tomo_1 = NXtomo()
    with pytest.raises(TypeError):
        nx_tomo_1.bliss_original_files = 12

    nx_tomo_1.bliss_original_files = ("/path/1", "/path/2")

    nx_tomo_2 = NXtomo()
    nx_tomo_2.bliss_original_files = ("/path/2", "/path/3")

    nx_tomo_3 = NXtomo()

    nx_tomo_4 = NXtomo()
    nx_tomo_4.bliss_original_files = ()

    nx_tomo_concat = concatenate([nx_tomo_1, nx_tomo_2, nx_tomo_3])
    assert nx_tomo_concat.bliss_original_files == ("/path/1", "/path/2", "/path/3")

    output_nx_tomo_concat = os.path.join(test_dir, "nx_concat.nx")
    nx_tomo_concat.save(output_nx_tomo_concat, "/entry_concat")

    loaded_nx_tomo = NXtomo().load(output_nx_tomo_concat, "/entry_concat")
    assert loaded_nx_tomo.bliss_original_files == ("/path/1", "/path/2", "/path/3")

    output_nx_tomo_file = os.path.join(test_dir, "nx_tomo.nx")
    nx_tomo_3.save(output_nx_tomo_file, "/entry0000")
    loaded_nx_tomo = NXtomo().load(output_nx_tomo_file, "/entry0000")
    assert loaded_nx_tomo.bliss_original_files is None

    nx_tomo_4.save(output_nx_tomo_file, "/entry0000", overwrite=True)
    loaded_nx_tomo = NXtomo().load(output_nx_tomo_file, "/entry0000")
    assert loaded_nx_tomo.bliss_original_files == ()


@pytest.mark.parametrize("vds_resolution", ("update", "remove"))
def test_copy_nxtomo_file(tmp_path, vds_resolution):
    """test 'copy_nxtomo_file' function"""
    input_folder = tmp_path / "input"
    input_folder.mkdir()
    input_nx_tomo_file = os.path.join(input_folder, "nexus.nx")

    output_folder = tmp_path / "output"
    output_folder.mkdir()

    nx_tomo = NXtomo()
    nx_tomo.save(input_nx_tomo_file, "/entry0000")

    output_file = os.path.join(output_folder, "nxtomo.nx")

    copy_nxtomo_file(
        input_nx_tomo_file,
        entries=None,
        output_file=output_file,
        vds_resolution=vds_resolution,
    )
    assert os.path.exists(output_file)


def test_multiple_readers(tmp_path):
    """Test that several readers can access the file in parallel with a thread pool or a process pool."""
    input_folder = tmp_path / "input"
    input_folder.mkdir()
    input_nx_tomo_file = os.path.join(input_folder, "nexus.nx")

    output_folder = tmp_path / "output"
    output_folder.mkdir()

    nx_tomo = NXtomo()
    detector_data = numpy.linspace(0, 100, 1000).reshape(10, 10, 10)
    nx_tomo.instrument.detector.data = detector_data
    nx_tomo.save(input_nx_tomo_file, "/entry0000")
    from time import sleep

    def read_data():
        with HDF5File(input_nx_tomo_file, mode="r") as h5f:
            # with h5py.File(input_nx_tomo_file, mode="r") as h5f:
            sleep(0.2)
            return h5f["/entry0000/instrument/detector/data"][()]

    futures = []
    with ThreadPoolExecutor(max_workers=1) as executor:
        for _ in range(10):
            futures.append(executor.submit(read_data))
    for future in futures:
        numpy.testing.assert_array_equal(future.result(), detector_data)

    with ProcessPoolExecutor() as executor:
        results = executor.map(read_data)
        for result in results:
            numpy.testing.assert_array_equal(result, detector_data)


@pytest.mark.skipif(not has_tifffile, reason="tifffile not installed")
@pytest.mark.parametrize("dtype", (numpy.uint16, numpy.float32))
@pytest.mark.parametrize("provide_dtype", (True, False))
@pytest.mark.parametrize("relative_link", (True, False))
def test_nxtomo_from_tiff(tmp_path, dtype, provide_dtype, relative_link):
    """Test creation of an NXtomo from a set of .tiff files."""
    tifffile_folder = tmp_path / "tiffs"
    tifffile_folder.mkdir()

    tiff_files = []
    raw_data = numpy.linspace(
        start=0,
        stop=1000,
        num=1000,
        dtype=dtype,
    ).reshape(10, 10, 10)
    for i in range(10):
        tiff_file = os.path.join(tifffile_folder, f"my_file{i}.tif")
        tifffile.imwrite(tiff_file, raw_data[i])
        tiff_files.append(tiff_file)

    output_nxtomo = os.path.join(tmp_path, "my_nxtomo.nx")
    nxtomo = NXtomo()
    with h5py.File(output_nxtomo, mode="w") as h5f:

        external_dataset_group = h5f.require_group("external_datasets")
        nxtomo.instrument.detector.data = create_detector_dataset_from_tiff(
            tiff_files=tiff_files,
            external_dataset_group=external_dataset_group,
            dtype=dtype if provide_dtype else None,
            relative_link=relative_link,
        )
    nxtomo.save(
        file_path=output_nxtomo,
        data_path="entry0000",
    )

    loaded_nxtomo = NXtomo().load(
        output_nxtomo, "entry0000", detector_data_as="as_numpy_array"
    )
    numpy.testing.assert_array_equal(
        loaded_nxtomo.instrument.detector.data,
        raw_data,
    )



# -- Project information -----------------------------------------------------

project = "nxtomo"
copyright = "2023-2025, ESRF"
author = "P.Paleo, H.Payno, A. Mirone, J.Lesaint, P.-O. Autran"

# The full version, including alpha/beta/rc tags
release = "3.0-dev"
version = release


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
    "sphinx.ext.doctest",
    "sphinx.ext.inheritance_diagram",
    "sphinx.ext.autosummary",
    "nbsphinx",
    "sphinx_autodoc_typehints",
    "myst_parser",
]

source_suffix = {
    ".rst": "restructuredtext",
    ".md": "markdown",
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pydata_sphinx_theme"

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

html_logo = "img/nxtomo.png"

# autosummary options
autosummary_generate = True
autosummary_imported_members = True

autodoc_default_flags = [
    "members",
    "undoc-members",
    "show-inheritance",
]


html_theme_options = {
    "icon_links": [
        {
            "name": "pypi",
            "url": "https://pypi.org/project/nxtomo",
            "icon": "_static/navbar_icons/pypi.svg",
            "type": "local",
        },
        {
            "name": "gitlab",
            "url": "https://gitlab.esrf.fr/tomotools/nxtomo",
            "icon": "_static/navbar_icons/gitlab.svg",
            "type": "local",
        },
    ],
    "show_toc_level": 1,
    "navbar_align": "left",
    "show_version_warning_banner": True,
    "navbar_start": ["navbar-logo", "version"],
    "navbar_center": ["navbar-nav"],
    "footer_start": ["copyright"],
    "footer_center": ["sphinx-version"],
}



"""NeXus paths used to define an `NXdetector <https://manual.nexusformat.org/classes/base_classes/NXdetector.html>`_."""


class NEXUS_DETECTOR_PATH:
    DATA = "data"

    IMAGE_KEY_CONTROL = "image_key_control"

    IMAGE_KEY = "image_key"

    X_PIXEL_SIZE = "x_pixel_size"

    Y_PIXEL_SIZE = "y_pixel_size"

    X_PIXEL_SIZE_MAGNIFIED = "x_magnified_pixel_size"

    Y_PIXEL_SIZE_MAGNIFIED = "y_magnified_pixel_size"

    X_REAL_PIXEL_SIZE = "real_x_pixel_size"

    Y_REAL_PIXEL_SIZE = "real_y_pixel_size"

    MAGNIFICATION = "magnification"

    DISTANCE = "distance"

    FOV = "field_of_view"

    ESTIMATED_COR_FRM_MOTOR = "estimated_cor_from_motor"
    "warning: replace by Y_ROTATION_AXIS_PIXEL_POSITION"

    ROI = "roi"

    EXPOSURE_TIME = "count_time"

    X_FLIPPED = "x_flipped"

    Y_FLIPPED = "y_flipped"

    NX_TRANSFORMATIONS = None
    # path in the NXdetector where are store the transformations

    X_ROTATION_AXIS_PIXEL_POSITION = None

    Y_ROTATION_AXIS_PIXEL_POSITION = None

    SEQUENCE_NUMBER = None


class NEXUS_DETECTOR_PATH_V_1_0(NEXUS_DETECTOR_PATH):
    pass


class NEXUS_DETECTOR_PATH_V_1_1(NEXUS_DETECTOR_PATH):
    pass


class NEXUS_DETECTOR_PATH_V_1_2(NEXUS_DETECTOR_PATH_V_1_1):
    pass


class NEXUS_DETECTOR_PATH_V_1_3(NEXUS_DETECTOR_PATH_V_1_2):
    # in this version we expect `x_flipped`, `y_flipped` to be replaced by  ̀TRANSFORMATIONS` NXtransformations group
    NX_TRANSFORMATIONS = "transformations"

    X_FLIPPED = None

    Y_FLIPPED = None


class NEXUS_DETECTOR_PATH_V_1_4(NEXUS_DETECTOR_PATH_V_1_3):

    ESTIMATED_COR_FRM_MOTOR = None  # replaced by 'X_ROTATION_AXIS_PIXEL_POSITION'

    X_ROTATION_AXIS_PIXEL_POSITION = "x_rotation_axis_pixel_position"

    Y_ROTATION_AXIS_PIXEL_POSITION = "y_rotation_axis_pixel_position"


class NEXUS_DETECTOR_PATH_V_1_5(NEXUS_DETECTOR_PATH_V_1_4):

    SEQUENCE_NUMBER = "sequence_number"
